# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.
""" Requests mock for Resilient REST API """

import logging
from collections import namedtuple
import json
import re
import requests_mock
from six import add_metaclass
LOG = logging.getLogger(__name__)
LOG.addHandler(logging.StreamHandler())
LOG.setLevel(logging.DEBUG)


def resilient_endpoint(request_type, uri):
    def mark(func):
        func.uri = uri
        func.request_type = request_type
        return func
    return mark


class ResilientMockType(type):
    def __new__(mcl, name, bases, nmspc):
        Endpoint = namedtuple("Endpoint", "type uri")
        try:
            endpoints = bases[0].registered_endpoints
        except:
            endpoints = {}
        for obj in nmspc.values():
            if hasattr(obj, 'uri'):
                endpoints[Endpoint(type=obj.request_type, uri=obj.uri)] = obj
        nmspc['registered_endpoints'] = endpoints
        return super(ResilientMockType, mcl).__new__(mcl, name, bases, nmspc)

@add_metaclass(ResilientMockType)
class ResilientMock(object):
    """ Base class for creating Resilient Rest API Mock definitions """

    def __init__(self, org_name=None, email=None):
        self.email = email or "api@example.com"
        self.org_name = org_name or "Test Org"
        LOG.info("Initialize ResilientMock %s %s", self.email, self.org_name)

        self.adapter = requests_mock.Adapter()
        for endpoint, handler in self.registered_endpoints.items():
            # Register with regex since some endpoints embed the org_id in the path
            LOG.debug("Registering %s %s to %s", endpoint.type,
                      endpoint.uri, str(handler))
            self.adapter.add_matcher(lambda request,
                                     method=endpoint.type,
                                     callback=handler,
                                     uri=endpoint.uri: self._custom_matcher(method,
                                                                   uri,
                                                                   lambda request: callback(self, request),
                                                                   request))

    @staticmethod
    def _custom_matcher(request_type, uri, response_callback, request):
        """ matcher function for passing to adapter.add_matcher() """
        if request.method == request_type and re.search(uri, request.url):
            return response_callback(request)
        else:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Common Helper Functions for the resilient library"""

import copy
import io
import json
import logging
import os
import shutil
import sys

from resilient import constants

if sys.version_info.major < 3:
    # Handle PY 2 specific imports
    from urllib import unquote

    from urlparse import urlparse
    JSONDecodeError = None  # JSONDecodeError is not available in PY2.7 so we set it to None
else:
    # Handle PY 3 specific imports
    from json.decoder import JSONDecodeError
    from urllib.parse import unquote, urlparse
    from jwcrypto import jwk, jwe


LOG = logging.getLogger(__name__)

CHARS_TO_MASK = [("?", "%3F"), ("#", "%23"), ("/", "%2F")]
MASK = "_*_{0}_*_"


def str_to_bool(value):
    """
    Convert value to either a ``True`` or ``False`` boolean

    Returns ``False`` if ``value`` is anything
    other than: ``'1', 'true', 'yes' or 'on'``

    :param value: the value to convert
    :type value: str
    :return: ``True`` or ``False``
    :rtype: bool
    """
    value = str(value).lower()
    return value in ('1', 'true', 'yes', 'on')


def mask_special_chars(s):
    """
    For any chars in CHARS_TO_MASK found in s
    replace them with their mask: _*_<URL Encoding minus the %>_*_

    Example:
    ```
    s = 'http://mockusername:mockpw%23%24%25%5E%26%2A%28%29-%2B_%3F%60%7E@192.168.0.5:3128'
    s = mask_special_chars(s)
    print(s)
    >>> 'http://mockusername:mockpw_*_23_*_%24%25%5E%26%2A%28%29-%2B__*_3F_*_%60%7E@192.168.0.5:3128'
    ```
    :param s: The string a urlencoded sting that may have characters to mask
    :type s: str
    :return: s replacing and special chars with their respective mask
    :rtype: str
    """
    if not s:
        return ""

    for c in CHARS_TO_MASK:
        mask = MASK.format(c[1][1:])
        s = s.replace(c[1], mask)
    return s


def unmask_special_chars(s):
    """
    Find any masked chars in s and replace them with their
    equivlent original character
    Example:
    ```
    s = 'mockpw_*_23_*_$%^&*()_*_3F_*_`~'
    s = unmask_special_chars(s)
    print(s)
    >>> 'mockpw#$%^&*()?`~'
    ```
    :param s: The string wanting to unmask
    :type s: str
    :return: s replaced with original chars if any
    :rtype: str
    """
    if not s:
        return ""

    for c in CHARS_TO_MASK:
        mask = MASK.format(c[1][1:])
        s = s.replace(mask, c[0])
    return s


def is_env_proxies_set():
    """
    :return: True/False if HTTP_PROXY or HTTPS_PROXY has a value
    :rtype: bool
    """
    if os.getenv(constants.ENV_HTTPS_PROXY) or os.getenv(constants.ENV_HTTP_PROXY):
        return True

    return False


def unquote_str(s):
    """
    Returns a string with the %xx replaced with their single-character
    equivalent

    If `s` None or an empty str, will just return an empty str

    :param s: String you want to unquote
    :type s: str

    :return: `s` with all %xx character replaced
    :rtype: str
    """
    if not s:
        return ""

    return unquote(s)


def get_and_parse_proxy_env_var(var_to_get=constants.ENV_HTTP_PROXY):
    """
    Get the `var_to_get` environment variable,
    and parse it returning a dictionary with the attributes:
    `scheme`, `hostname`, `port`, `username` and `password`

    `username` and `password` will be empty strings if not provided in the var

    `var_to_get` is by default HTTP_PROXY

    :param var_to_get: A str of the name of the env var to get and parse
    :type var_to_get: str
    :return: a dict included attributes with information of the proxy
    :rtype: dict
    """

    var = os.getenv(var_to_get)

    if not var:
        return {}

    var = mask_special_chars(var)
    var = unquote_str(var)
    parsed_var = urlparse(var)

    return {
        "scheme": parsed_var.scheme,
        "hostname": parsed_var.hostname,
        "port": parsed_var.port,
        "username": unmask_special_chars(parsed_var.username),
        "password": unmask_special_chars(parsed_var.password)
    }


def is_in_no_proxy(host, no_proxy_var=constants.ENV_NO_PROXY):
    """
    Return True if `host` is found in `no_proxy_var`
    else returns False

    :param host: must be a str and a fully qualified domain name or an IPv4 Address
    :type host: str
    :return: a bool whether or not `host` is in NO_PROXY env var
    :rtype: bool
    """
    if not host or not isinstance(host, str):
        return False

    no_proxy = os.getenv(no_proxy_var)

    if not no_proxy or not isinstance(no_proxy, str):
        return False

    if host in no_proxy:
        return True

    return False


def remove_tag(original_res_obj):
    """
    Return the original_res_obj with any of the "tags"
    attribute set to an empty list

    Example:
    ```
    mock_res_obj = {
        "tags": [{"tag_handle": "fn_tag_test", "value": None}],
        "functions": [
            {"export_key": "fn_tag_test_function",
            "tags": [{'tag_handle': 'fn_tag_test', 'value': None}]}
        ]
    }

    new_res_obj = remove_tag(mock_res_obj)

    Returns: {
        "tags": [],
        "functions": [
            {"export_key": "fn_tag_test_function", "tags": []}
        ]
    }
    ```
    :param original_res_obj: the res_obj you want to remove the tags attribute from
    :type original_res_obj: dict
    :return: new_res_obj: a dict with the tag attribute removed
    :rtype: dict
    """
    ATTRIBUTE_NAME = "tags"

    new_res_obj = copy.deepcopy(original_res_obj)

    if isinstance(new_res_obj, dict):

        # Set "tags" to empty list
        if new_res_obj.get(ATTRIBUTE_NAME):
            new_res_obj[ATTRIBUTE_NAME] = []

        # Recursively loop the dict
        for obj_name, obj_value in new_res_obj.items():

            if isinstance(obj_value, list):
                for index, obj in enumerate(obj_value):
                    new_res_obj[obj_name][index] = remove_tag(obj)

            elif isinstance(obj_value, dict):
                new_res_obj[obj_name] = remove_tag(obj_value)

    return new_res_obj


def is_running_in_app_host(env_var=constants.ENV_VAR_APP_HOST_CONTAINER):
    """
    Checks if the APP_HOST_CONTAINER environmental variable
    is set

    :param env_var: name of the APP_HOST_CONTAINER environmental variable, defaults to constants.ENV_VAR_APP_HOST_CONTAINER
    :type env_var: str, optional
    :return: True if it is set to 1, else False
    :rtype: bool
    """
    if not str_to_bool(get_config_from_env(env_var)):
        LOG.debug("Not running in an App Host environment")
        return False

    return True


def protected_secret_exists(secret_name, path_secrets_dir=constants.PATH_SECRETS_DIR, path_jwk_file=constants.PATH_JWK_FILE):
    """
    Check to see if the APP_HOST_CONTAINER env var is set, the /etc/secrets directory,
    the SECRET_FILE with the encrypted token and the key.jwk file all exist and
    the user has the correct permissions to read them

    :param secret_name:  Name of the protected secret file
    :type secret_name: str
    :param path_secrets_dir: Path to the location of the encrypted secret files, defaults to constants.PATH_SECRETS_DIR, defaults to constants.PATH_SECRETS_DIR
    :type path_secrets_dir: str, optional
    :param path_jwk_file: Path to the location of the jwk.key file in a JSON format as per https://www.ietf.org/rfc/rfc7517.txt, defaults to constants.PATH_JWK_FILE
    :type path_jwk_file: str
    :return: True if all files are found and the user has the correct permission, False otherwise
    :rtype: bool
    """
    path_secret = os.path.join(path_secrets_dir, secret_name)

    if not is_running_in_app_host():
        return False

    if sys.version_info.major < 3:
        LOG.warning(constants.WARNING_PROTECTED_SECRETS_NOT_SUPPORTED)
        return False

    if not os.path.isdir(path_secrets_dir) or not os.access(path_secrets_dir, os.R_OK):
        LOG.debug("Protected secrets directory at '%s' does not exist or you do not have the correct permissions. No value found for '%s'", path_secrets_dir, secret_name)
        return False

    if not os.path.isfile(path_secret) or not os.access(path_secret, os.R_OK):
        LOG.warning("No protected secret found for '%s' or you do not have the correct permissions to read the file. No value found for '%s'", secret_name, secret_name)
        return False

    if not os.path.isfile(path_jwk_file) or not os.access(path_jwk_file, os.R_OK):
        LOG.warning("Could not find JWK at '%s' or you do not have the correct permissions. No value found for '%s'", path_jwk_file, secret_name)
        return False

    return True


def get_protected_secret(secret_name, path_secrets_dir=constants.PATH_SECRETS_DIR, path_jwk_file=constants.PATH_JWK_FILE):
    """
    Get the JWK, read the token from a file with
    the secret_name and decrypt it using the JWK

    :param secret_name: Name of the protected secret file
    :type secret_name: str
    :param path_secrets_dir: Path to the location of the encrypted secret files, defaults to constants.PATH_SECRETS_DIR
    :type path_secrets_dir: str, optional
    :param path_jwk_file: Path to the location of the jwk.key file in a JSON format as per https://www.ietf.org/rfc/rfc7517.txt, defaults to constants.PATH_JWK_FILE
    :type path_jwk_file: str
    :return: The decrypted value of the protected secret
    :rtype: str
    """
    LOG.info("Reading Protected Secret '%s'", secret_name)

    if sys.version_info.major < 3:
        LOG.warning(constants.WARNING_PROTECTED_SECRETS_NOT_SUPPORTED)
        return None

    path_secret = os.path.join(path_secrets_dir, secret_name)
    tkn = None
    key = get_jwk(path_jwk_file)

    if not key:
        return None

    with io.open(path_secret, mode="r", encoding="utf-8") as f:
        tkn = f.readline()

    if not tkn:
        LOG.error("File for protected secret '%s' is empty or corrupt", secret_name)
        return None

    # We need to remove new line and carriage return characters
    tkn = tkn.splitlines()[0]

    try:
        jwetoken = jwe.JWE()
        jwetoken.deserialize(tkn)
        jwetoken.decrypt(key)
        decrypted_value = jwetoken.payload
    except Exception as err:
        LOG.error("Could not decrypt the secret. Invalid key used to decrypt the protected secret '%s'. Error Message: %s", secret_name, str(err))
        return None

    return decrypted_value.decode("utf-8")


def get_jwk(path_jwk_file=constants.PATH_JWK_FILE):
    """
    If the contents of the file at path is valid JSON,
    reads the file and uses the jwcrypto.jwk.JWK class
    get the JWK and returns it else returns None

    :param path_jwk_file: Path to JSON JWK file to read
    :type path_jwk_file: str
    :return: File contents as a jwcrypto.jwk.JWK or None
    :rtype: jwcrypto.jwk.JWK
    """
    LOG.debug("Getting JWK from '%s'", path_jwk_file)

    file_contents = None

    if not os.path.isfile(path_jwk_file) or not os.access(path_jwk_file, os.R_OK):
        LOG.warning("Could not find JWK at '%s' or you do not have the correct permissions.", path_jwk_file)
        return None

    with io.open(path_jwk_file, mode="rt", encoding="utf-8") as the_file:

        try:
            file_contents = json.load(the_file)

        except JSONDecodeError as err:
            LOG.error("JWK JSON file at '%s' is corrupt.\njwk: %s\nError: %s", path_jwk_file, jwk, str(err))
            return None

    if not file_contents:
        LOG.error("The provided JWK file at '%s' is empty", path_jwk_file)
        return None

    try:
        aes_key = jwk.JWK.from_json(json.dumps(file_contents))

    except Exception as err:
        LOG.error("Error getting JWK: %s", str(err))
        return None

    return aes_key


def remove_secrets_dir(path_secrets_dir=constants.PATH_SECRETS_DIR):
    """
    Check if we running in App Host and if the secrets directory
    exists, remove it

    :param path_secrets_dir: Path to the location of the encrypted secret files, defaults to constants.PATH_SECRETS_DIR
    :type path_secrets_dir: str, optional
    """
    if is_running_in_app_host() and os.path.isdir(path_secrets_dir):
        LOG.debug("Removing secrets directory at: '%s'", path_secrets_dir)
        shutil.rmtree(path_secrets_dir, ignore_errors=True)


def get_config_from_env(config_name):
    """
    Read a variable from the environment given it's
    config_name. If it does not exist, it returns None

    :param config_name: Name of the env var to get
    :type config_name: str
    :return: The value of the env var
    :rtype: str
    """
    LOG.debug("Getting environmental variable '%s'", config_name)
    return os.environ.get(config_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import logging
import os
import sys

import pytest
from resilient import constants, helpers

if sys.version_info.major >= 3:
    # Handle PY 3 specific imports
    from jwcrypto.jwk import JWK


def test_str_to_bool():
    assert helpers.str_to_bool("True") is True
    assert helpers.str_to_bool("true") is True
    assert helpers.str_to_bool("1") is True
    assert helpers.str_to_bool("YeS") is True
    assert helpers.str_to_bool("ON") is True
    assert helpers.str_to_bool("2") is False
    assert helpers.str_to_bool(1) is True
    assert helpers.str_to_bool(0) is False
    assert helpers.str_to_bool("0") is False
    assert helpers.str_to_bool(None) is False


def test_is_env_proxies_set(fx_add_proxy_env_var):
    assert helpers.is_env_proxies_set() is True


def test_is_env_proxies_not_set():
    assert helpers.is_env_proxies_set() is False


def test_unquote_str():
    assert helpers.unquote_str("""%C2%B1%21%40%23%24%25%5E%26%2A%28%29_%2B""") == """±!@#$%^&*()_+"""
    assert helpers.unquote_str("""mockusername""") == """mockusername"""
    assert helpers.unquote_str("""mockpw1234567890%21%40%23%24%25%5E%26%2A%28%29-%2B_%3D%5B%5D%7B%3B%27%5C%7D%3A%22%7C/.%2C%3C%3E%3F%60%7E""") == """mockpw1234567890!@#$%^&*()-+_=[]{;'\}:"|/.,<>?`~"""
    assert helpers.unquote_str("") == ""
    assert helpers.unquote_str(None) == ""


def test_get_and_parse_proxy_env_var_not_set():
    assert helpers.get_and_parse_proxy_env_var() == {}


def test_get_and_parse_proxy_env_var(fx_add_proxy_env_var):
    proxy_details = helpers.get_and_parse_proxy_env_var()
    assert proxy_details["scheme"] == "http"
    assert proxy_details["hostname"] == "192.168.0.5"
    assert proxy_details["port"] == 3128
    assert proxy_details["username"] == ""
    assert proxy_details["password"] == ""


def test_get_and_parse_proxy_env_var_with_user(fx_add_proxy_env_var_with_user):
    proxy_details = helpers.get_and_parse_proxy_env_var()
    assert proxy_details["scheme"] == "http"
    assert proxy_details["hostname"] == "192.168.0.5"
    assert proxy_details["port"] == 3128
    assert proxy_details["username"] == "mockusername"
    assert proxy_details["password"] == """mockpw1234567890!@@@@#$%^&*()-+_=[]{;\'\\}:"/|,<>?`~"""


def test_is_in_no_proxy(fx_add_proxy_env_var):
    assert helpers.is_in_no_proxy("subdomain.example.com") is True
    assert helpers.is_in_no_proxy("192.168.1.5") is True
    assert helpers.is_in_no_proxy("domain.not.in.com") is False
    assert helpers.is_in_no_proxy("") is False
    assert helpers.is_in_no_proxy(None) is False


def test_no_proxy_is_not_set():
    assert helpers.is_in_no_proxy("subdomain.example.com") is False
    assert helpers.is_in_no_proxy("192.168.1.5") is False
    assert helpers.is_in_no_proxy("domain.not.in.com") is False
    assert helpers.is_in_no_proxy("") is False
    assert helpers.is_in_no_proxy(None) is False


def test_remove_tag():

    mock_res_obj = {
        "tags": [{"tag_handle": "fn_tag_test", "value": None}],
        "functions": [
            {"export_key": "fn_tag_test_function", "tags": [{'tag_handle': 'fn_tag_test', 'value': None}]}
        ],
        "workflows": {
            "nested_2": [{"export_key": "fn_tag_test_function", "tags": [{'tag_handle': 'fn_tag_test', 'value': None}]}]
        }
    }

    new_res_obj = helpers.remove_tag(mock_res_obj)

    assert new_res_obj.get("tags") == []
    assert new_res_obj.get("functions", [])[0].get("tags") == []
    assert new_res_obj.get("workflows", []).get("nested_2")[0].get("tags") == []


def test_is_running_in_app_host(fx_reset_environmental_variables):
    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"
    assert helpers.is_running_in_app_host() is True


def test_is_not_running_in_app_host(caplog):
    caplog.set_level(logging.DEBUG)
    assert helpers.is_running_in_app_host() is False
    assert "Not running in an App Host environment" in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_protected_secret_exists(fx_write_protected_secrets, fx_reset_environmental_variables):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")
    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"

    assert helpers.protected_secret_exists("API_KEY", path_secrets_dir, path_jwk_file) is True
    assert helpers.protected_secret_exists("API_KEY", "invalid_path", path_jwk_file) is False
    assert helpers.protected_secret_exists("INVALID_SECRET_NAME", path_secrets_dir, path_jwk_file) is False
    assert helpers.protected_secret_exists("API_KEY", path_secrets_dir, "invalid_path") is False


@pytest.mark.skipif(sys.version_info >= constants.MIN_SUPPORTED_PY3_VERSION, reason="only run this test in Python 2.7")
def test_protected_secret_exists_unsupported_python_version(fx_reset_environmental_variables, caplog):

    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"
    assert helpers.protected_secret_exists("API_KEY", "mock_path", "mock_path") is False
    assert "Protected secrets are only Python >= 3 supported" in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_protected_secret_exists_env_var_not_set(fx_write_protected_secrets):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")

    assert helpers.protected_secret_exists("API_KEY", path_secrets_dir, path_jwk_file) is False


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_protected_secret(fx_write_protected_secrets, caplog):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")

    assert helpers.get_protected_secret("API_KEY", path_secrets_dir, path_jwk_file) == "JbkOxTInUg1aIRGxXI8zOG1A25opU39lDKP1_0rfeVQ"


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_protected_secret_empty_file(fx_write_protected_secrets, caplog):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")

    assert helpers.get_protected_secret("EMPTY", path_secrets_dir, path_jwk_file) is None
    assert "File for protected secret 'EMPTY' is empty or corrupt" in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_protected_secret_wrong_key(fx_write_protected_secrets, caplog):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key_unused.jwk")

    assert helpers.get_protected_secret("API_KEY", path_secrets_dir, path_jwk_file) is None
    assert "Could not decrypt the secret. Invalid key used to decrypt the protected secret 'API_KEY'." in caplog.text


@pytest.mark.skipif(sys.version_info >= constants.MIN_SUPPORTED_PY3_VERSION, reason="only run this test in Python 2.7")
def test_get_protected_secret_unsupported_python_version(caplog):

    assert helpers.get_protected_secret("API_KEY", "mock_path", "mock_path") is None
    assert "Protected secrets are only Python >= 3 supported" in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_jwk(fx_write_protected_secrets):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")

    jwk = helpers.get_jwk(path_jwk_file)
    in_json = jwk.export(as_dict=True)

    assert isinstance(jwk, JWK)
    assert in_json.get("k") == "W4VMVYqZg-xGiDTGjFFwRvbzf098AskheVdZPbpiYvE"


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_jwk_no_key(fx_write_protected_secrets, caplog):
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, "EMPTY")

    assert helpers.get_jwk(path_jwk_file) is None
    assert "JWK JSON file at '{0}' is corrupt.".format(path_jwk_file) in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_get_jwk_invalid_file_path(fx_write_protected_secrets, caplog):
    assert helpers.get_jwk("invalid_path") is None
    assert "Could not find JWK at 'invalid_path' or you do not have the correct permissions." in caplog.text


def test_get_config_from_env(fx_reset_environmental_variables):
    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"
    assert helpers.get_config_from_env(constants.ENV_VAR_APP_HOST_CONTAINER) == "1"
    assert helpers.get_config_from_env("invalid_env_var") is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""
Shared pytest fixtures

Note:
    -   Code after the 'yield' statement in a fixture
        is ran after the test (or scope i.e. test session) has complete
    -   fx_ prefixes a 'fixture'
    -   Put fixture logic in separate 'private' function so we
        can share logic between fixtures
    -   Fixture must have BEFORE and AFTER docstring
"""

import os
import shutil
import sys
import copy

import pytest
import requests_mock
from resilient import constants
from resilient.co3 import SimpleClient
from resilient.co3base import BaseClient

from tests.shared_mock_data import mock_paths


def _mk_temp_dir():
    if os.path.exists(mock_paths.TEST_TEMP_DIR):
        shutil.rmtree(mock_paths.TEST_TEMP_DIR)

    os.makedirs(mock_paths.TEST_TEMP_DIR)


def _rm_temp_dir():
    if os.path.exists(mock_paths.TEST_TEMP_DIR):
        shutil.rmtree(mock_paths.TEST_TEMP_DIR, ignore_errors=True)


def _add_to_cmd_line_args(args_to_add):
    sys.argv.extend(args_to_add)


@pytest.fixture
def fx_base_client():
    """
    Before: Creates sample BaseClient object with mocked URI https://example.com
    After: Nothing
    """
    base_client = BaseClient(org_name="Mock Org", base_url="https://example.com")
    requests_adapter = requests_mock.Adapter()
    base_client.session.mount(u'https://', requests_adapter)
    base_client.org_id = 201
    yield (base_client, requests_adapter)


@pytest.fixture
def fx_simple_client():
    """
    Before: Creates sample SimpleClient object with mocked URI https://example.com
    After: Nothing
    """
    base_client = SimpleClient(org_name="Mock Org", base_url="https://example.com")
    requests_adapter = requests_mock.Adapter()
    base_client.session.mount(u'https://', requests_adapter)
    base_client.org_id = 201
    yield (base_client, requests_adapter)


@pytest.fixture
def fx_mk_temp_dir():
    """
    Before: Creates a directory at mock_paths.TEST_TEMP_DIR
    After: Removes the directory
    """
    _mk_temp_dir()
    yield
    _rm_temp_dir()


@pytest.fixture
def fx_add_proxy_env_var():
    """
    Before: set HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    After: unset HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    """
    os.environ[constants.ENV_HTTPS_PROXY] = "https://192.168.0.5:3128"
    os.environ[constants.ENV_HTTP_PROXY] = "http://192.168.0.5:3128"
    os.environ[constants.ENV_NO_PROXY] = "subdomain.example.com,192.168.1.5,.example.com"

    yield

    os.environ[constants.ENV_HTTPS_PROXY] = ""
    os.environ[constants.ENV_HTTP_PROXY] = ""
    os.environ[constants.ENV_NO_PROXY] = ""


@pytest.fixture
def fx_add_proxy_env_var_with_user():
    """
    Before: set HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    After: unset HTTPS_PROXY, HTTP_PROXY and NO_PROXY
    """
    os.environ[constants.ENV_HTTPS_PROXY] = "https://mockusername:mockpw1234567890%21%40%23%24%25%5E%26%2A%28%29-%2B_%3D%5B%5D%7B%3B%27%5C%7D%3A%22%2F%7C%2C%3C%3E%3F%60%7E@192.168.0.5:3128"
    os.environ[constants.ENV_HTTP_PROXY] = "http://mockusername:mockpw1234567890%21%40%40%40%40%23%24%25%5E%26%2A%28%29-%2B_%3D%5B%5D%7B%3B%27%5C%7D%3A%22%2F%7C%2C%3C%3E%3F%60%7E@192.168.0.5:3128"
    os.environ[constants.ENV_NO_PROXY] = "subdomain.example.com,192.168.1.5,.example.com"

    yield

    os.environ[constants.ENV_HTTPS_PROXY] = ""
    os.environ[constants.ENV_HTTP_PROXY] = ""
    os.environ[constants.ENV_NO_PROXY] = ""


@pytest.fixture
def fx_write_protected_secrets(fx_mk_temp_dir):
    """
    Before: Create temp directory using fx_mk_temp_dir fixture and copy mock_secrets to it
    After: Remove the dir
    """
    path_secrets_dir = os.path.join(mock_paths.TEST_TEMP_DIR, os.path.basename(mock_paths.MOCK_SECRETS_DIR))
    shutil.copytree(mock_paths.MOCK_SECRETS_DIR, path_secrets_dir)
    yield path_secrets_dir


@pytest.fixture
def fx_reset_environmental_variables():
    """
    Before: Create a deepcopy of current env variables
    After: Set the current env variables back to their original

    Used in a test where we want to modify the env vars and avoid leaking into other tests
    """
    current_env = copy.deepcopy(os.environ)

    yield

    os.environ = current_env



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

"""Utility to access schema metadata with the Resilient REST API"""

from __future__ import print_function
import resilient
import json
import sys
import codecs
import csv
import collections
import logging

if sys.version_info.major < 3:
    from StringIO import StringIO
else:
    from io import StringIO


logging.basicConfig(format='%(asctime)s %(message)s', level=logging.WARN)
LOG = logging.getLogger(__name__)


def wrap_io(stream):
    """Wrap the stream to always always output in utf-8"""
    if stream.encoding != 'UTF-8':
        if sys.version_info.major < 3:
            return codecs.getwriter('utf-8')(stream, 'strict')
        else:
            return codecs.getwriter('utf-8')(stream.buffer, 'strict')
    return stream

# Always always output in utf-8
sys.stdout = wrap_io(sys.stdout)
sys.stderr = wrap_io(sys.stderr)


class FinfoArgumentParser(resilient.ArgumentParser):
    def __init__(self, config_file=None):
        super(FinfoArgumentParser, self).__init__(config_file=config_file)

        self.add_argument('fieldname',
                          nargs="?",
                          help="The field name.")

        self.add_argument('--types',
                          action='store_true',
                          help="Print the list of types.")

        self.add_argument('--type',
                          dest="field_type",
                          default="incident",
                          help="The object type.  This can be the API Access Name of a "
                               "Data Table, or one of the standard types: 'incident' "
                               "(the default), 'task', 'artifact', 'milestone', "
                               "'attachment', 'note', or 'actioninvocation'.")

        self.add_argument('--json',
                          action='store_true',
                          help="Print the field definition in JSON format.")

        self.add_argument('--csv',
                          action='store_true',
                          help="Print the field lists in CSV format.")

        self.add_argument('--values',
                          dest="field_values",
                          action='store_true',
                          help="Print the list of valid values for all fields.")


def apiname(field):
    """The full (qualified) programmatic name of a field"""
    if field["prefix"]:
        fieldname = u"{}.{}".format(field["prefix"], field["name"])
    else:
        fieldname = field["name"]
    return fieldname


def print_json(field):
    """Print the definition of one field, in JSON"""
    print(json.dumps(field, indent=4))


def print_details(field):
    """Print the definition of one field, in readable text"""
    print(u"Name:        {}".format(apiname(field)))
    print(u"Label:       {}".format(field["text"]))
    print(u"ID:          {}".format(field["id"]))
    print(u"Type:        {}".format(field["input_type"]))
    if "tooltip" in field:
        if field["tooltip"]:
            print(u"Tooltip:     {}".format(field["tooltip"]))
    if "placeholder" in field:
        if field["placeholder"]:
            print(u"Placeholder: {}".format(field["placeholder"]))
    if "required" in field:
        print(u"Required:    {}".format(field["required"]))
    if "values" in field:
        if field["values"]:
            print("Values:")
            v = sorted(field["values"], key=lambda x: x["value"])
            for value in v:
                default_flag = " "
                if value["default"]:
                    default_flag = "*"
                if not value["enabled"]:
                    default_flag = "x"
                label = value["label"]
                print (u'{} {}={}'.format(default_flag, value["value"], label))


def find_field(client, fieldname, objecttype="incident"):
    trimname = fieldname[fieldname.rfind(".")+1:]
    t = client.get("/types/{}/fields".format(objecttype))
    for field in t:
        if field["name"] == trimname:
            return field


def list_fields_csv(client, objecttype="incident"):
    """Print a list of fields, in CSV format"""
    iostr = StringIO()
    writer = None
    t = client.get("/types/{}/fields".format(objecttype))
    for field in sorted(t, key=apiname):
        columns = collections.OrderedDict()
        columns["name"] = apiname(field)
        columns["required"] = field.get("required", "")
        columns["input_type"] = field.get("input_type", "")

        if sys.version_info.major < 3:
            columns["text"] = field.get("text", "").encode("utf-8")
            columns["tooltip"] = field.get("tooltip", "").encode("utf-8")
            columns["placeholder"] = field.get("placeholder", "").encode("utf-8")
        else:
            columns["text"] = field.get("text", "")
            columns["tooltip"] = field.get("tooltip", "")
            columns["placeholder"] = field.get("placeholder", "")

        if not writer:
            writer = csv.DictWriter(iostr, fieldnames=columns.keys(), dialect='excel')
            writer.writeheader()
        writer.writerow(columns)

    result = iostr.getvalue()
    if sys.version_info.major < 3:
        print(result.decode("utf-8"))
    else:
        print(result)


def list_fields_values(client, objecttype="incident"):
    """Print a list of fields' valid values, in CSV format"""
    LOG.info("values")
    iostr = StringIO()
    writer = None
    t = client.get("/types/{}/fields".format(objecttype))
    for field in sorted(t, key=apiname):
        if field.get("input_type", "") not in ["multiselect_members", "select_owner"]:
            if "values" in field:
                if field["values"]:
                    v = sorted(field["values"], key=lambda x: x["value"])
                    for value in v:
                        e = ""
                        if not value["enabled"]:
                            e = "False"
                        columns = collections.OrderedDict()
                        columns["name"] = apiname(field)
                        columns["enabled"] = e
                        columns["value"] = value["value"]

                        if sys.version_info.major < 3:
                            columns["default"] = str(value["default"] or "").encode("utf-8")
                            columns["label"] = (value["label"] or "").encode("utf-8")
                        else:
                            columns["default"] = str(value["default"] or "")
                            columns["label"] = str(value["label"] or "")

                        if not writer:
                            writer = csv.DictWriter(iostr, fieldnames=columns.keys(), dialect='excel')
                            writer.writeheader()
                        writer.writerow(columns)

    result = iostr.getvalue()
    if sys.version_info.major < 3:
        print(result.decode("utf-8"))
    else:
        print(result)


def list_fields(client, objecttype="incident"):
    """Print a list of fields, in readable text"""
    print("Fields:")
    t = client.get("/types/{}/fields".format(objecttype))
    for field in sorted(t, key=apiname):
        required_flag = " "
        if "required" in field:
            if field["required"] == "always":
                required_flag = "*"
            if field["required"] == "close":
                required_flag = "c"
        print(u"{} {}".format(required_flag, apiname(field)))


def list_types(client):
    """Print a list of types, in readable text"""
    print("Types:")
    t = client.get("/types")

    def print_types(parent, indent):
        if indent > 10:
            return
        for type in sorted(t.keys()):
            typedef = t[type]
            typename = typedef["display_name"]
            parents = typedef["parent_types"]
            if (parent is None and len(parents) == 0) or (parent in parents):
                print(u"{}{}".format(' ' * indent, type))
                print_types(type, indent+2)

    print_types(None, 2)


def main():
    """Main"""
    # Parse commandline arguments
    parser = FinfoArgumentParser(config_file=resilient.get_config_file())
    opts = parser.parse_args()

    # Connect to Resilient
    client = resilient.get_client(opts)

    # If no field is specified, list them all
    if not opts.fieldname:
        if opts.types:
            list_types(client)
        elif opts.field_values:
            list_fields_values(client, opts.field_type)
        elif opts.csv:
            list_fields_csv(client, opts.field_type)
        else:
            list_fields(client, opts.field_type)
        exit(0)

    # Find the field and display its properties
    field_data = find_field(client, opts.fieldname, opts.field_type)
    if field_data:
        if opts.json:
            print_json(field_data)
        else:
            print_details(field_data)
        exit(0)
    else:
        print(u"Field '{}' was not found.".format(opts.fieldname))
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""
    res-utils.py
    ----------

    This is a utility tool resilient SDK. Customers can
    1. Verify the connection to a resilient server
    2. Create/get incidents
    3. Attach files
    4. Get type information
    5. Get action information
    6. ...
"""
import resilient
import argparse
import logging
import json
import subprocess
import requests

import tempfile
import os

class ResClient(object):
    """

    """
    def __init__(self):
        self.res_client = None
        self.other_args = None

    def connect(self):
        print("----------------")
        print("Ready to connect")
        print("----------------")
        print("Read config information from app.confg ...")
        arg_parser = resilient.ArgumentParser(resilient.get_config_file()).parse_args(args=self.other_args)
        host = arg_parser.host
        email = arg_parser.email
        password = arg_parser.password
        org = arg_parser.org
        api_key_id = arg_parser.api_key_id
        api_key_secret = arg_parser.api_key_secret

        cafile = arg_parser.cafile

        verify = True
        if cafile is not None and cafile == "false":
            verify = False
        url = "https://{}:443".format(host)
        print("Connecting to {} using user:{}, and org:{}".format(url, email, org))
        print("Validate cert: {}".format(verify))

        args = {"base_url": url,
                "verify": verify,
                "org_name": org}

        self.res_client = resilient.SimpleClient(**args)

        if email is not None and password is not None:
            session = self.res_client.connect(email, password)

            if session is not None:
                user_name = session.get("user_displayname", "Not found")
                print("User display name is : {}".format(user_name))

            print("Done")
        else:
            self.res_client.set_api_key(api_key_id=api_key_id,
                                        api_key_secret=api_key_secret)

    def create_incident(self, infile):
        self.connect()
        with open(infile, "r") as infile:
            text = infile.read()
            print("Trying to post this dict as incident: {}".format(text))
            inc_dict = json.loads(text)
            self.res_client.post("/incidents", inc_dict)

    def count_incidents(self):
        self.connect()
        incidents = self.query_incidents()
        print("There are {} incidents".format(len(incidents)))

    def incident_type(self):
        self.connect()
        type_dict = self.res_client.get("/types/incident/fields")
        print("Incident type/fields: {}".format(type_dict))

    def get_actions(self):
        self.connect()
        resp = self.res_client.get("/actions")
        print("Actions: {}".format(resp))

    def query_incidents(self, max_count=None, page_size=1000, in_log=None):
        print("----------------------------")
        print("Download incidents and count")
        print("----------------------------")

        log = in_log if in_log else logging.getLogger(__name__)
        incidents = []
        url = "/incidents/query_paged?field_handle=-1&return_level=full"
        num_incidents = 0
        ret_num = 0
        done = False
        while not done:
            body = {
                "start": num_incidents,
                "length": page_size,
                "recordsTotal": page_size
            }
            ret = self.res_client.post(uri=url,
                                       payload=body)

            data = ret.get("data", [])
            ret_num = len(data)
            if ret_num > 0:
                log.debug("Downloaded {} incidents, total now {} ...".format(ret_num, ret_num + num_incidents))
                incidents.extend(data)
            else:
                #
                # No more to read.
                #
                done = True

            num_incidents = num_incidents + ret_num

            if max_count:
                if num_incidents >= max_count:
                    #
                    # Reach max_count set by user, stop now
                    #
                    done = True

        return incidents

    def attach_to_incident(self, filename, inc_id):
        print("-----------------------------------------")
        print("Attaching file {} to incident {}".format(filename, inc_id))
        print("-----------------------------------------")
        self.connect()

        basename = os.path.basename(filename)
        attachment_uri = "/incidents/{0}/attachments".format(inc_id)
        new_attachment = self.res_client.post_attachment(attachment_uri, filename,
                                                         filename=basename)
        print(new_attachment)

    def attach_to_incident_artifact(self, filename, inc_id, type):
        print("-----------------------------------------")
        print("Attaching file {} to incident {} as an artifact (type {})".format(filename, inc_id, type))
        print("-----------------------------------------")
        self.connect()

        basename = os.path.basename(filename)
        attachment_uri = "/incidents/{0}/artifacts/files".format(inc_id)

        new_attachment = self.res_client.post_artifact_file(uri=attachment_uri,
                                                            artifact_type=int(type),
                                                            artifact_filepath=filename,
                                                            description="Description",
                                                            value=basename)
        print(new_attachment)

def check_connect():
    """
    Use openssl and python requests to check the connection with Resilient
    :return:
    """
    arg_parser = resilient.ArgumentParser(resilient.get_config_file())
    host = arg_parser.getopt("resilient", "host")
    #
    # Use Openssl first
    #
    print("-------------------------------------")
    print("Using openssl to connect to resilient")
    print("-------------------------------------")
    command = "openssl s_client -connect {}:443".format(host)
    user = arg_parser.getopt("resilient", "email")
    password = arg_parser.getopt("resilient", "password")
    process = subprocess.Popen("/bin/bash", stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = process.communicate(command)
    cafile = arg_parser.getopt("resilient", "cafile")
    verify = True
    if cafile is not None and cafile == "false":
        verify = False
    print(out)
    if err is not None:
        print(err)

    print("---------------------------------------------")
    print("Using python requests to connect to resilient")
    print("---------------------------------------------")

    rest_url = "https://{}:443/rest/session".format(host)
    data = '{"email": "' + user + '","password":"' + password + '", "interactive": true}'
    try:
        header = {"Content-Type": "application/json"}
        resp = requests.post(rest_url,
                             data=data,
                             headers=header,
                             verify=verify)
        print("\tResponse: " + str(resp))

    except Exception as e:
        print("\tConnection failed!!")
        print("\t" + str(e))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="resilient SDK utility tool")

    subparsers = parser.add_subparsers(title="subcommands",
                                       help="one of these subcommands must be provided",
                                       description="valid subcommands",
                                       dest="cmd")
    subparsers.required = True

    conn_parser = subparsers.add_parser("connect",
                                        help="Make a connection")

    count_parser = subparsers.add_parser("count",
                                         help="Count incidents")

    create_parser = subparsers.add_parser("create",
                                          help="Create a new incident")

    create_parser.add_argument("-i", "--input",
                               help="file with json dict for creating a new incident",
                               required=True)

    incident_type_parser = subparsers.add_parser("incident_type",
                                                 help="Get incident type")

    check_connect_parser = subparsers.add_parser("check_connect",
                                                 help="Check connect using openssl/requests/sdk")

    get_actions_parser = subparsers.add_parser("get_actions",
                                               help="List actions")

    # attach file to incident
    attach_parser = subparsers.add_parser("attach",
                                          help="Attach a file to an incident")
    attach_parser.add_argument("-f", "--file",
                               help="file name of file to attach")
    attach_parser.add_argument("-i", "--incident",
                               help="incident id")

    # attach file to incident
    attach_artifact_parser = subparsers.add_parser("attach_artifact",
                                                   help="Attach a file to an incident as an artifact")
    attach_artifact_parser.add_argument("-f", "--file",
                                        help="file name of file to attach")
    attach_artifact_parser.add_argument("-i", "--incident",
                                        help="incident id")
    attach_artifact_parser.add_argument("-t", "--type",
                                        help="artifact type")

    args, unknown_args = parser.parse_known_args()

    res_client = ResClient()
    res_client.other_args = unknown_args

    if args.cmd == "connect":
        res_client.connect()
    elif args.cmd == "count":
        res_client.count_incidents()
    elif args.cmd == "create":
        infile = args.input
        res_client.create_incident(infile)
    elif args.cmd == "incident_type":
        res_client.incident_type()
    elif args.cmd == "check_connect":
        check_connect()
        res_client.connect()
    elif args.cmd == "get_actions":
        res_client.connect()
        res_client.get_actions()
    elif args.cmd == "attach":
        filename = args.file
        inc_id = args.incident
        res_client.attach_to_incident(filename, inc_id)
    elif args.cmd == "attach_artifact":
        filename = args.file
        inc_id = args.incident
        type = args.type
        res_client.attach_to_incident_artifact(filename, inc_id, type)





# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""Base client for Resilient REST API"""
from __future__ import print_function

import json
import logging
import mimetypes
import os
import ssl
import sys
import traceback
import unicodedata

import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
from requests.packages.urllib3.poolmanager import PoolManager
from requests_toolbelt.multipart.encoder import MultipartEncoder
from retry.api import retry_call

from resilient import constants, helpers

try:
    # Python 3
    import urllib.parse as urlparse
except:
    # Python 2
    import urlparse

LOG = logging.getLogger(__name__)


class TLSHttpAdapter(HTTPAdapter):
    """
    Adapter that ensures that we use the best available SSL/TLS version.
    Some environments default to SSLv3, so we need to specifically ask for
    the highest protocol version that both the client and server support.
    """
    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(num_pools=connections,
                                       maxsize=maxsize,
                                       block=block,
                                       ssl_version=ssl.PROTOCOL_TLS if sys.version_info.major == 2 else ssl.PROTOCOL_TLS_CLIENT)


class BasicHTTPException(Exception):
    """Exception for HTTP errors."""
    def __init__(self, response, err_reason=u"Unknown Reason", err_text=u"Unknown Error"):
        """
        Args:
          response - the Response object from the get/put/etc.
        """

        err_reason = response.reason if response.reason else err_reason
        err_text = response.text if response.text else err_text

        err_message = u"'resilient' API Request FAILED:\nResponse Code: {0}\nReason: {1}. {2}".format(response.status_code, err_reason, err_text)

        # Add a __qualname__ attribute if does not exist - needed for PY27
        if not hasattr(BasicHTTPException, "__qualname__"):
            setattr(BasicHTTPException, "__qualname__", BasicHTTPException.__name__)

        super(BasicHTTPException, self).__init__(err_message)

        self.response = response

    def get_response(self):
        return self.response

    @staticmethod
    def raise_if_error(response):

        if response.status_code == 401:
            try:
                raise BasicHTTPException(response, err_reason=constants.ERROR_MSG_CONNECTION_UNAUTHORIZED, err_text=constants.ERROR_MSG_CONNECTION_INVALID_CREDS)
            except BasicHTTPException:
                traceback.print_exc()
                sys.exit(constants.ERROR_CODE_CONNECTION_UNAUTHORIZED)

        elif response.status_code != 200:
            raise BasicHTTPException(response)


class NoChange(Exception):
    """
    Exception that can be raised within a get/put handler or a patch callback
    to indicate 'no change' (which then just bypasses the update operation).
    """
    pass


def ensure_unicode(input_value):
    """ if input_value is type str, convert to unicode with utf-8 encoding """
    if sys.version_info.major >= 3:
        return input_value

    if not isinstance(input_value, basestring):
        return input_value
    if isinstance(input_value, str):
        input_unicode = input_value.decode('utf-8')
    else:
        input_unicode = input_value

    input_unicode = unicodedata.normalize('NFKC', input_unicode)
    return input_unicode


def get_proxy_dict(opts):
    """ Creates a dictionary with proxy config to be sent to the SimpleClient """
    scheme = urlparse.urlparse(opts['proxy_host']).scheme
    if not scheme:
        scheme = 'https'
        proxy_host = opts['proxy_host']
    else:
        proxy_host = opts['proxy_host'][len(scheme + "://"):]

    if opts.get('proxy_user') and opts.get('proxy_password'):
        proxy = {'https': '{0}://{1}:{2}@{3}:{4}/'.format(scheme, opts['proxy_user'], opts['proxy_password'],
                                                          proxy_host, opts['proxy_port'])}
    else:
        proxy = {'https': '{0}://{1}:{2}'.format(scheme, proxy_host, opts['proxy_port'])}

    return proxy


class BaseClient(object):
    """Helper for using SOAR REST API."""

    def __init__(self, org_name=None, base_url=None, proxies=None, verify=None, certauth=None, custom_headers=None, **kwargs):
        """
        :param org_name: The name of the organization to use
        :type org_name: str
        :param base_url: The base URL of the SOAR server, e.g. ``https://soar.ibm.com/``
        :type base_url: str
        :param proxies: A dictionary of ``HTTP`` proxies to use, if any
        :type proxies: dict
        :param verify: The path to a ``PEM`` file containing the trusted CAs, or ``False`` to disable all TLS verification
        :type verify: str|bool
        :param certauth: The filepath for the client side certificate and the private key either as a single file or as a tuple of both files' paths
        :type certauth: str|tuple(str, str)
        :param custom_headers: A dictionary of any headers you want to send in **every** request
        :type custom_headers: dict
        :param kwargs: A dictionary of any other keyword arguments
        :type kwargs: dict
        """

        base_headers = {
            "content-type": "application/json",
            constants.HEADER_MODULE_VER_KEY: constants.HEADER_MODULE_VER_VALUE
        }

        if custom_headers and isinstance(custom_headers, dict):
            base_headers.update(custom_headers)

        self.headers = base_headers

        self.cookies = None
        self.org_id = None
        self.user_id = None
        self.base_url = u'https://app.resilientsystems.com/'
        self.org_name = ensure_unicode(org_name)
        if proxies:
            self.proxies = {ensure_unicode(key): ensure_unicode(proxies[key]) for key in proxies}
        else:
            self.proxies = None
        if helpers.is_env_proxies_set():
            self.proxies = None
        if base_url:
            self.base_url = ensure_unicode(base_url)
        self.verify = verify
        self.verify = ensure_unicode(verify)
        if verify is None:
            self.verify = True
        self.authdata = None
        self.session = requests.Session()
        self.session.mount(u'https://', TLSHttpAdapter())

        # API key
        self.api_key_id = None
        self.api_key_secret = None
        self.use_api_key = False
        self.api_key_handle = None      # This is the principle ID for an api key. Also called handle

        # Client Cert based authentication
        self.cert = certauth

        # Retry configs
        self.max_connection_retries = kwargs.get(constants.APP_CONFIG_MAX_CONNECTION_RETRIES) if kwargs.get(constants.APP_CONFIG_MAX_CONNECTION_RETRIES) is not None else constants.APP_CONFIG_MAX_CONNECTION_RETRIES_DEFAULT
        self.request_max_retries = kwargs.get(constants.APP_CONFIG_REQUEST_MAX_RETRIES) if kwargs.get(constants.APP_CONFIG_REQUEST_MAX_RETRIES) is not None else constants.APP_CONFIG_REQUEST_MAX_RETRIES_DEFAULT
        self.request_retry_delay = kwargs.get(constants.APP_CONFIG_REQUEST_RETRY_DELAY) if kwargs.get(constants.APP_CONFIG_REQUEST_RETRY_DELAY) is not None else constants.APP_CONFIG_REQUEST_RETRY_DELAY_DEFAULT
        self.request_retry_backoff = kwargs.get(constants.APP_CONFIG_REQUEST_RETRY_BACKOFF) if kwargs.get(constants.APP_CONFIG_REQUEST_RETRY_BACKOFF) is not None else constants.APP_CONFIG_REQUEST_RETRY_BACKOFF_DEFAULT

    def set_api_key(self, api_key_id, api_key_secret, timeout=None, include_permissions=False):
        """
        Call this method instead of the connect method in order to use API key
        Just like the connect method, this method calls the session endpoint
        to get org_id information.
        :param api_key_id: api key ID to use to connect
        :type api_key_id: string
        :param api_key_secret: associated secret
        :type api_key_secret: string
        :param timeout: timeout limit if desired. None by default
        :type timeout: float
        :param include_permissions: whether to include permissions in call to /rest/session.
            Since SOAR v48 this param has been included and set to "true" by default on the server
            (until v50 where it will be removed). We don't need permission details in circuits so we
            set it to False by default, but if there is a use of this elsewhere in app code,
            and either "perms" or "effective_permissions" details that are returned by the
            endpoint are needed, the value here should be set to True.
        :type include_permissions: bool
        :return:
        """
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.use_api_key = True

        if include_permissions:
            LOG.debug("'include_permissions' is deprecated and scheduled to be removed in v50, use GET " +
                        "/rest/session/{org_id}/acl instead.\n\t\tAt that time, 'include_permissions' will be " +
                        "removed and this endpoint will not return org permissions.")

        # Wrap self.session.get and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __set_api_key():
            r = self.session.get(u"{0}/rest/session?include_permissions={1}".format(self.base_url, "true" if include_permissions else "false"),
                                 auth=HTTPBasicAuth(self.api_key_id, self.api_key_secret),
                                 proxies=self.proxies,
                                 headers=self.make_headers(),
                                 verify=self.verify,
                                 timeout=timeout,
                                 cert=self.cert)
            BasicHTTPException.raise_if_error(r)
            return r

        response = retry_call(__set_api_key,
                              exceptions=(BasicHTTPException, ConnectionError),
                              tries=self.max_connection_retries,
                              delay=self.request_retry_delay,
                              backoff=self.request_retry_backoff)

        session = json.loads(response.text)
        self._extract_org_id(session)
        self.api_key_handle = session.get("api_key_handle", None)
        return session

    def connect(self, email, password, timeout=None):
        """Performs connection, which includes authentication.

        Args:
          email - the email address to use for authentication.
          password - the password
          timeout - number of seconds to wait for response
        Returns:
          The Resilient session object (dict)
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """
        self.authdata = {
            u'email': ensure_unicode(email),
            u'password': ensure_unicode(password)
        }
        return self._connect(timeout=timeout)

    def _extract_org_id(self, response):
        """
        Extract the org's id from server response. Loops the orgs
        in the response and checks each ``cloud_account``, ``uuid``
        or ``name`` against the ``org_name`` attribute in the 
        app.config file to see if it matches any of the 3.

        If found, sets ``self.org_id`` equal to the ``id`` of the org

        Also sets ``self.all_orgs`` equal to the full list of orgs
        that are ``enabled`` for that user

        :param response: server response from session endpoint
        :type response: dict
        :raises Exception: if user is not a member of any orgs
        :raises Exception: if ``org_name`` value in app.config is not set
        :raises Exception: if the user is not a member of the org specified in the app.config
        :raises Exception: if the org specified has an ``enabled`` state set to ``False``
        """
        app_config_org_value = self.org_name
        selected_org = None
        orgs = response.get("orgs", [])

        if not orgs:
            raise Exception("User is not a member of any orgs")

        if not app_config_org_value:
            org_names = [org.get("name") for org in orgs]
            msg = u"Please specify the organization name to which you want to connect.\n" + \
                  u"The user is a member of the following organizations: '{0}'"
            raise Exception(msg.format(u"', '".join(org_names)))

        for o in orgs:

            if app_config_org_value == o.get("cloud_account", ""):
                LOG.info("Using cloud account id: %s", app_config_org_value)
                selected_org = o
                break

            if app_config_org_value == o.get("uuid", ""):
                LOG.info("Using org uuid: %s", app_config_org_value)
                selected_org = o
                break

            if ensure_unicode(app_config_org_value) == o.get("name", ""):
                LOG.info("Using org name: %s", app_config_org_value)
                selected_org = o
                break

        if not selected_org:
            msg = u"The user is not a member of the specified organization '{0}'."
            raise Exception(msg.format(self.org_name))

        if not selected_org.get("enabled", False):
            msg = "This organization is not accessible to you.\n\n" \
                  "This can occur because of one of the following:\n\n" \
                  "The organization does not allow access from your current IP address.\n" \
                  "The organization requires authentication with a different provider than you are currently using.\n" \
                  "Your IP address is {0}"
            raise Exception(msg.format(response.get("session_ip", "Unknown")))

        self.all_orgs = [org for org in orgs if org.get("enabled")]
        self.org_id = selected_org.get("id", None)

    def _connect(self, timeout=None, include_permissions=False):
        """Connect to SOAR using deprecated username and password method"""

        if include_permissions:
            LOG.debug("'include_permissions' is deprecated and scheduled to be removed in v50, use GET " +
                        "/rest/session/{org_id}/acl instead.\n\t\tAt that time, 'include_permissions' will be " +
                        "removed and this endpoint will not return org permissions.")

        # Wrap self.session.post and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __connect():
            r = self.session.post(u"{0}/rest/session?include_permissions={1}".format(self.base_url, "true" if include_permissions else "false"),
                                  data=json.dumps(self.authdata),
                                  proxies=self.proxies,
                                  headers=self.make_headers(),
                                  verify=self.verify,
                                  timeout=timeout,
                                  cert=self.cert)

            BasicHTTPException.raise_if_error(r)
            return r

        response = retry_call(__connect,
                              exceptions=(BasicHTTPException, ConnectionError),
                              tries=self.max_connection_retries,
                              delay=self.request_retry_delay,
                              backoff=self.request_retry_backoff)

        session = json.loads(response.text)
        self._extract_org_id(session)

        # set the X-sess-id token, which is used to prevent CSRF attacks.
        self.headers['X-sess-id'] = session['csrf_token']
        self.cookies = {
            'JSESSIONID': response.cookies['JSESSIONID']
        }
        self.user_id = session["user_id"]
        return session

    def make_headers(self, co3_context_token=None, additional_headers=None):
        """Makes a headers dict, including the X-Co3ContextToken (if co3_context_token is specified)."""
        headers = self.headers.copy()

        if co3_context_token is not None:
            headers['X-Co3ContextToken'] = co3_context_token
        if isinstance(additional_headers, dict):
            headers.update(additional_headers)
        return headers

    def _execute_request(self, operation, url, **kwargs):
        """
        If self.use_api_key is set to ``True``, set the ``auth`` header
        of the request to a ``HTTPBasicAuth`` using the api key id and secret

        Then execute the request

        :param operation: the requests.Session method to call
        :type operation: requests.Session.get | requests.Session.post | requests.Session.delete
        :param url: URL to send request to
        :type url: str
        :return: requests.response object
        :rtype: requests.response
        """

        if self.use_api_key:
            kwargs["auth"] = HTTPBasicAuth(self.api_key_id, self.api_key_secret)
            #
            #   Note this is a temporary walk around. Theoretically the server
            #   shall ignore the session id if api key is used. But we don't have
            #   time to do that yet. So we clear the session id here
            #
            self.session.cookies.clear()

        result = operation(url, **kwargs)

        return result

    def get(self, uri, co3_context_token=None, timeout=None, is_uri_absolute=None, get_response_object=None):
        """Gets the specified URI.  Note that this URI is relative to <base_url>/rest/orgs/<org_id>.  So
        for example, if you specify a uri of /incidents, the actual URL would be something like this:

            https://app.resilientsystems.com/rest/orgs/201/incidents

        Args:
          uri
          co3_context_token
          timeout: number of seconds to wait for response
          is_uri_absolute: if True, does not insert /org/{org_id} into the uri
          get_response_object: if True, returns the response object rather than the json of the response.text
        Returns:
          A dictionary, array, or response object with the value returned by the server.
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """

        if is_uri_absolute:
            url = u"{0}/rest{1}".format(self.base_url, ensure_unicode(uri))

        else:
            url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))

        # Wrap _execute_request and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __get():
            r = self._execute_request(self.session.get,
                                      url,
                                      proxies=self.proxies,
                                      cookies=self.cookies,
                                      headers=self.make_headers(co3_context_token),
                                      verify=self.verify,
                                      timeout=timeout,
                                      cert=self.cert)
            BasicHTTPException.raise_if_error(r)
            return r

        response = retry_call(__get,
                              exceptions=(BasicHTTPException, ConnectionError),
                              tries=self.request_max_retries,
                              delay=self.request_retry_delay,
                              backoff=self.request_retry_backoff)

        if get_response_object:
            return response

        return json.loads(response.text)

    def get_const(self, co3_context_token=None, timeout=None):
        """
        Get the ``ConstREST`` endpoint.

        Endpoint for retrieving various constant information for this server. This information is
        useful in translating names that the user sees to IDs that other REST API endpoints accept.

        For example, the ``incidentDTO`` has a field called ``"crimestatus_id"``. The valid values are stored
        in ``constDTO.crime_statuses``.

        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: ``ConstDTO`` as a dictionary
        :rtype: dict
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        url = u"{0}/rest/const".format(self.base_url)
        response = self._execute_request(self.session.get,
                                         url,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token),
                                         verify=self.verify,
                                         timeout=timeout)
        BasicHTTPException.raise_if_error(response)
        return response.json()

    def get_content(self, uri, co3_context_token=None, timeout=None):
        """Gets the specified URI.  Note that this URI is relative to <base_url>/rest/orgs/<org_id>.  So
        for example, if you specify a uri of /incidents, the actual URL would be something like this:

            https://app.resilientsystems.com/rest/orgs/201/incidents

        Args:
          uri
          co3_context_token
          timeout: number of seconds to wait for response
        Returns:
          The raw value returned by the server for this resource.
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))
        response = self._execute_request(self.session.get,
                                         url,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token),
                                         verify=self.verify,
                                         timeout=timeout,
                                         cert=self.cert)
        BasicHTTPException.raise_if_error(response)
        return response.content

    def post(self, uri, payload, co3_context_token=None, timeout=None, headers=None):
        """
        Posts to the specified URI.
        Note that this URI is relative to <base_url>/rest/orgs/<org_id>.  So for example, if you
        specify a uri of /incidents, the actual URL would be something like this:

            https://app.resilientsystems.com/rest/orgs/201/incidents
        Args:
           uri
           payload
           co3_context_token
           timeout: number of seconds to wait for response
           headers: optional headers to include
        Returns:
          A dictionary or array with the value returned by the server.
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))

        # payloads which aren't convertable to json are passed asis
        payload_json = json.dumps(payload) if isinstance(payload, (list, dict)) else payload

        #----
        # Wrap _execute_request and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __post():
            r = self._execute_request(self.session.post,
                                      url,
                                      data=payload_json,
                                      proxies=self.proxies,
                                      cookies=self.cookies,
                                      headers=self.make_headers(co3_context_token, additional_headers=headers),
                                      verify=self.verify,
                                      timeout=timeout,
                                      cert=self.cert)
            BasicHTTPException.raise_if_error(r)
            return r

        response = retry_call(__post,
                              exceptions=(BasicHTTPException, ConnectionError),
                              tries=self.request_max_retries,
                              delay=self.request_retry_delay,
                              backoff=self.request_retry_backoff)

        BasicHTTPException.raise_if_error(response)
        try:
            return json.loads(response.text)
        except json.decoder.JSONDecodeError:
            return response.content

    def post_attachment(self, uri, filepath,
                        filename=None,
                        mimetype=None,
                        data=None,
                        co3_context_token=None,
                        timeout=None,
                        bytes_handle=None):
        """
        Upload a file to the specified URI
        e.g. "/incidents/<id>/attachments" (for incident attachments)
        or,  "/tasks/<id>/attachments" (for task attachments)

        :param uri: The REST URI for posting
        :param filepath:the path of the file to post or if ``None``, use ``bytes_handle``
        :param filename: optional name of the file when posted
        :param mimetype: optional override for the guessed MIME type
        :param data: optional dict with additional MIME parts (not required for file attachments; used in artifacts)
        :param co3_context_token: Action Module context token, if responding to an Action Module event
        :param timeout: optional timeout (seconds)
        :param bytes_handle: BytesIO handle for content, used if ``filepath`` is None
        """
        filepath = ensure_unicode(filepath)
        if filename:
            filename = ensure_unicode(filename)
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))
        mime_type = mimetype or mimetypes.guess_type(filename or filepath)[0] or "application/octet-stream"

        # Wrap _execute_request and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __post_attachment(name=None, file_or_bytes_handle=None, extra_data=None):
            multipart_data = {'file': (name, file_or_bytes_handle, mime_type)}
            multipart_data.update(extra_data or {})
            encoder = MultipartEncoder(fields=multipart_data)
            headers = self.make_headers(co3_context_token, additional_headers={'content-type': encoder.content_type})
            r = self._execute_request(self.session.post,
                                      url,
                                      data=encoder,
                                      proxies=self.proxies,
                                      cookies=self.cookies,
                                      headers=headers,
                                      verify=self.verify,
                                      timeout=timeout,
                                      cert=self.cert)
            BasicHTTPException.raise_if_error(r)
            return json.loads(r.text)

        if filepath:
            with open(filepath, 'rb') as file_handle:
                attachment_name = filename or os.path.basename(filepath)
                json_resp = retry_call(__post_attachment,
                                       fkwargs={
                                           "name": attachment_name,
                                           "file_or_bytes_handle": file_handle,
                                           "extra_data": data
                                       },
                                       exceptions=(BasicHTTPException, ConnectionError),
                                       tries=self.request_max_retries,
                                       delay=self.request_retry_delay,
                                       backoff=self.request_retry_backoff)
                return json_resp

        elif bytes_handle:
            attachment_name = filename if filename else "Unknown"
            json_resp = retry_call(__post_attachment,
                                   fkwargs={
                                       "name": attachment_name,
                                       "file_or_bytes_handle": bytes_handle,
                                       "extra_data": data
                                   },
                                   exceptions=(BasicHTTPException, ConnectionError),
                                   tries=self.request_max_retries,
                                   delay=self.request_retry_delay,
                                   backoff=self.request_retry_backoff)
            return json_resp

        else:
            raise ValueError("Either filepath or bytes_handle are required")

    def post_artifact_file(self, uri, artifact_type, artifact_filepath,
                           description=None,
                           value=None,
                           mimetype=None,
                           co3_context_token=None,
                           timeout=None,
                           bytes_handle=None):
        """
        Post a file artifact to the specified URI
        e.g. "/incidents/<id>/artifacts/files"

        :param uri: The REST URI for posting
        :param artifact_type: the artifact type name ("IP Address", etc) or type ID
        :param artifact_filepath: the path of the file to post or ``None`` if using ``bytes_handle``
        :param description: optional description for the artifact
        :param value: optional value for the artifact
        :param mimetype: optional override for the guessed MIME type
        :param co3_context_token: Action Module context token, if responding to an Action Module event
        :param timeout: optional timeout (seconds)
        :param bytes_handle: byte content to create as an artifact file, used if ``artifact_filepath`` is ``None``

        """
        artifact = {
            "type": artifact_type,
            "value": value or "",
            "description": description or ""
        }
        mimedata = {
            "artifact": json.dumps(artifact)
        }
        return self.post_attachment(uri,
                                    artifact_filepath,
                                    filename=value if bytes_handle else None,
                                    mimetype=mimetype,
                                    data=mimedata,
                                    co3_context_token=co3_context_token,
                                    timeout=timeout,
                                    bytes_handle=bytes_handle)

    def _get_put(self, uri, apply_func, co3_context_token=None, timeout=None):
        """Internal helper to do a get/apply/put loop
        (for situations where the put might return a 409/conflict status code)
        """
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))
        response = self._execute_request(self.session.get,
                                         url,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token),
                                         verify=self.verify,
                                         timeout=timeout,
                                         cert=self.cert)
        BasicHTTPException.raise_if_error(response)
        payload = json.loads(response.text)
        try:
            apply_func(payload)
        except NoChange:
            return payload
        payload_json = json.dumps(payload)
        response = self._execute_request(self.session.put,
                                         url,
                                         data=payload_json,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token),
                                         verify=self.verify,
                                         timeout=timeout,
                                         cert=self.cert)
        if response.status_code == 200:
            return json.loads(response.text)
        if response.status_code == 409:
            return None
        BasicHTTPException.raise_if_error(response)
        return None

    def get_put(self, uri, apply_func, co3_context_token=None, timeout=None):
        """Performs a get, calls apply_func on the returned value, then calls self.put.
        If the put call returns a 409 error, then retry.

        Args:
          uri - the URI to use.  Note that this is expected to be relative to the org.
          apply_func - a function to call on the object returned by get.  This is expected
          to alter the object with the desired changes.
          co3_context_token - the Co3ContextToken from a CAF message (if the caller is
          a CAF message processor.
          timeout - number of seconds to wait for response
        Returns;
          The object returned by the put operation (converted from JSON to a Python dict).
        Raises:
          Exception if the get or put returns an unexpected status code.
        """
        while True:
            obj = self._get_put(uri, apply_func, co3_context_token=co3_context_token, timeout=timeout)
            if obj:
                return obj
        return None

    def put(self, uri, payload, co3_context_token=None, timeout=None, headers=None):
        """
        Puts to the specified URI.
        Note that this URI is relative to <base_url>/rest/orgs/<org_id>.  So for example, if you
        specify a uri of /incidents, the actual URL would be something like this:

            https://app.resilientsystems.com/rest/orgs/201/incidents
        Args:
           uri
           payload
           headers: optional headers to include
           co3_context_token
           timeout: number of seconds to wait for response
        Returns:
          A dictionary or array with the value returned by the server.
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))
        # payloads which aren't convertable to json are passed asis
        payload_json = json.dumps(payload) if isinstance(payload, (list, dict)) else payload
        response = self._execute_request(self.session.put,
                                         url,
                                         data=payload_json,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token, additional_headers=headers),
                                         verify=self.verify,
                                         timeout=timeout,
                                         cert=self.cert)
        BasicHTTPException.raise_if_error(response)
        return json.loads(response.text)

    def delete(self, uri, co3_context_token=None, timeout=None):
        """Deletes the specified URI.

        Args:
          uri
          co3_context_token
          timeout: number of seconds to wait for response
        Returns:
          A dictionary or array with the value returned by the server.
        Raises:
          BasicHTTPException - if an HTTP exception occurs.
        """
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))

        # Wrap _execute_request and its related raise_if_error call in
        # inner function so we can add retry logic with dynamic parameters to it
        def __delete():
            r = self._execute_request(self.session.delete,
                                      url,
                                      proxies=self.proxies,
                                      cookies=self.cookies,
                                      headers=self.make_headers(co3_context_token),
                                      verify=self.verify,
                                      timeout=timeout,
                                      cert=self.cert)
            if r.status_code == 204:
                # 204 - No content is OK for a delete
                return None

            BasicHTTPException.raise_if_error(r)
            return json.loads(r.text)

        response = retry_call(__delete,
                              exceptions=(BasicHTTPException, ConnectionError),
                              tries=self.request_max_retries,
                              delay=self.request_retry_delay,
                              backoff=self.request_retry_backoff)

        return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""Simple client for Resilient REST API"""
import datetime
import importlib
import json
import logging
import os
import sys
from argparse import Namespace

import requests
from cachetools import TTLCache, cachedmethod

from . import co3base
from .co3base import NoChange, ensure_unicode, get_proxy_dict
from .patch import PatchStatus

try:
    # Python 3
    import urllib.parse as urlparse
except:
    # Python 2
    import urlparse


DEFAULT_CONFIG_FILENAME = "app.config"
LOG = logging.getLogger(__name__)


def get_resilient_circuits_version():
    """
    If resilient-circuits is installed, returns its version
    Else will return None
    """
    try:
        import resilient_circuits
        res_circuits_version = resilient_circuits.__version__.split(".")
        return {
            "major": int(res_circuits_version[0]),
            "minor": int(res_circuits_version[1]),
            "patch": int(res_circuits_version[2])
        }

    except ImportError:
        pass

    return None


def get_config_file(filename=None, generate_filename=False):
    """
    Helper: get the location of the configuration file
    * Use the location specified in $APP_CONFIG_FILE, if set
    * Otherwise if filename path specified in args exist in the current working use as config file.
    * Otherwise if default config file name exists in current work directory use as config file.
    * Otherwise use path in ~/.resilient/ directory

    :param filename: The filename to use for the app config file.
    :param generate_filename: Boolean is used for config filename generation.
    """
    # The config file location should usually be set in the environment
    # First check environment, then cwd, then ~/.resilient/app.config
    env_app_config_file = os.environ.get("APP_CONFIG_FILE", None)

    if not env_app_config_file:
        if generate_filename and filename:
            # If generating the config file and filename passed in, use it as the config file name.
            config_file = os.path.expandvars(os.path.expanduser(filename))
        else:
            if not filename:
                # If file not specified use default value.
                filename = DEFAULT_CONFIG_FILENAME
            # If the filename exists in local directory use as config file name.
            if not generate_filename and os.path.exists(filename):
                config_file = filename
            else:
                # Use default config file in ~/.resilient/app.config.
                config_file = os.path.expanduser(os.path.join("~", ".resilient", filename))
                if generate_filename:
                    # If generating the config file location, create the '~/.resilient' directory if missing.
                    resilient_dir = os.path.dirname(config_file)
                    if not os.path.exists(resilient_dir):
                        LOG.info(u"Creating %s", resilient_dir)
                        os.makedirs(resilient_dir)
    else:
        config_file = env_app_config_file
    return config_file


def get_client(opts, custom_headers=None, **kwargs):
    """
    This is a helper method to get an instance of
    :class:`SimpleClient` for the SOAR REST API.

    :param opts: the connection options - usually the contents of the app.config file
    :type opts: dict
    :param custom_headers: A dictionary of any headers you want to send in **every** request
    :type custom_headers: dict
    :param kwargs: A dictionary of any other keyword arguments
    :type kwargs: dict

    A standard way to initialize a :class:`SimpleClient` with default configuration is:

    .. code-block:: python

        parser = resilient.ArgumentParser(config_file=resilient.get_config_file())
        opts = parser.parse_args()
        client = resilient.get_client(opts)

    :return: a connected and verified instance of :class:`SimpleClient`
    """
    if isinstance(opts, Namespace):
        opts = vars(opts)

    # Allow explicit setting "do not verify certificates"
    verify = opts.get("cafile")
    if str(verify).lower() == "false":
        LOG.warning("Unverified HTTPS requests (cafile=false).")
        requests.packages.urllib3.disable_warnings()  # otherwise things get very noisy
        verify = False

    # Enable client certificate authentication
    clientauthcert = opts.get("client_auth_cert", None)
    clientauthkey = opts.get("client_auth_key", None)
    certauth = (clientauthcert, clientauthkey)
    if str(clientauthcert).lower() == "false" or clientauthcert is None or clientauthkey is None:
        certauth = False

    proxy = None
    if opts.get("proxy_host"):
        proxy = get_proxy_dict(opts)

    # Create SimpleClient for a REST connection to the Resilient services
    url = "https://{0}:{1}".format(opts.get("host", ""), opts.get("port", 443))
    url = urlparse.urljoin(url, opts.get("resource_prefix", ""))
    simple_client_args = {"org_name": opts.get("org"),
                          "proxies": proxy,
                          "base_url": url,
                          "verify": verify,
                          "certauth": certauth,
                          "custom_headers": custom_headers}
    if opts.get("log_http_responses"):
        LOG.warning("Logging all HTTP Responses from Resilient to %s", opts["log_http_responses"])
        simple_client = LoggingSimpleClient
        simple_client_args["logging_directory"] = opts["log_http_responses"]
    else:
        simple_client = SimpleClient

    # Update with kwargs
    simple_client_args.update(kwargs)

    resilient_client = simple_client(**simple_client_args)

    if opts.get("resilient_mock"):
        # Use a Mock for the Resilient Rest API
        LOG.warning("Using Mock '%s' for Resilient REST API", opts["resilient_mock"])
        module_path, class_name = opts["resilient_mock"].rsplit('.', 1)
        path, module_name = os.path.split(module_path)
        sys.path.insert(0, path)
        module = importlib.import_module(module_name)
        LOG.info("Looking for %s in %s", class_name, dir(module))
        mock_class = getattr(module, class_name)
        res_mock = mock_class(org_name=opts.get("org"), email=opts["email"])
        resilient_client.session.mount("https://", res_mock.adapter)

    #
    #   Check if we are using API Key or user/password. API key is the
    #   preferable one if we can find from the opts
    #
    if opts.get("api_key_id", None) is not None and opts.get("api_key_secret", None) is not None:
        userinfo = resilient_client.set_api_key(api_key_id=opts["api_key_id"],
                                                api_key_secret=opts["api_key_secret"])
    else:
        userinfo = resilient_client.connect(opts["email"], opts["password"])

    # Validate the org, and store org_id in the opts dictionary
    LOG.debug(json.dumps(userinfo, indent=2))
    if(len(userinfo["orgs"])) > 1 and opts.get("org") is None:
        raise Exception("User is a member of multiple organizations; please specify one.")
    if(len(userinfo["orgs"])) > 1:
        for org in userinfo["orgs"]:
            if org["name"] == opts.get("org"):
                opts["org_id"] = org["id"]
    else:
        opts["org_id"] = userinfo["orgs"][0]["id"]

    # Check if action module is enabled and store to opts dictionary
    org_data = resilient_client.get('')
    resilient_client.actions_enabled = org_data["actions_framework_enabled"]

    return resilient_client


class SimpleHTTPException(Exception):
    """Exception for HTTP errors."""
    def __init__(self, response):
        """
        :param response: The Response object from the get/put/etc.
        """
        super(SimpleHTTPException, self).__init__(u"{0}:  {1}".format(response.reason, response.text))

        self.response = response


class PatchConflictException(SimpleHTTPException):
    """Exception for patch conflicts."""
    def __init__(self, response, patch_status):
        super(PatchConflictException, self).__init__(response)

        self.patch_status = patch_status


def _raise_if_error(response):
    """Helper to raise a SimpleHTTPException if the response.status_code is not 200.

    :param response: the Response object from a get/put/etc.
    :raises SimpleHTTPException: if response.status_code is not 200.
    """
    if response.status_code != 200:
        raise SimpleHTTPException(response)


class SimpleClient(co3base.BaseClient):
    """
    Python helper class for using the IBM SOAR REST API.

    .. code-block::

        '''
        An example script showing how to connect and get
        incidents with IBM Security SOAR

        Example Usage:
        $ python sample_connect.py "https://<host>", "<org_name>", "<api_key_id>", "<api_key_secret>", "<path_to_ca_file>|False"
        '''

        import sys
        import os
        from resilient import SimpleClient


        def main():

            args = sys.argv
            assert len(args) == 6
            SCRIPT_NAME = args[0]

            # Get required parameters
            HOST = args[1]
            ORG_NAME = args[2]
            API_KEY_ID = args[3]
            API_KEY_SECRET = args[4]

            # Check if using cafile
            ca_file = args[5]
            VERIFY = ca_file if os.path.isfile(ca_file) is True else False

            # Instansiate SimpleClient
            res_client = SimpleClient(
                org_name=ORG_NAME,
                base_url=HOST,
                verify=VERIFY
            )

            # Set the API Key
            res_client.set_api_key(
                api_key_id=API_KEY_ID,
                api_key_secret=API_KEY_SECRET
            )

            # Setup the request payload
            payload = {
                "filters": [
                    {
                        "conditions": [{"field_name": "plan_status", "method": "equals", "value": "A"}]
                    }
                ]
            }

            # Invoke the request
            response = res_client.post("/incidents/query_paged?return_level=full", payload=payload)

            # Print results
            for incident in response.get("data", []):
                print("{0}: {1}".format(incident.get("id"), incident.get("name")))


        if __name__ == "__main__":
            main()


    .. note::
        The full REST API Documentation for IBM SOAR can be
        viewed on the Platform itself by searching for the **REST API Reference**
        or by going to: ``https://<base_url>/docs/rest-api/index.html``
    """

    def __init__(self, org_name=None, base_url=None, proxies=None, verify=None, cache_ttl=240, certauth=None, custom_headers=None, **kwargs):
        """
        :param org_name: The name of the organization to use.
        :type org_name: str
        :param base_url: The base URL of the SOAR server, e.g. ``https://soar.ibm.com/``
        :type base_url: str
        :param proxies: A dictionary of ``HTTP`` proxies to use, if any.
        :type proxies: dict
        :param verify: The path to a ``PEM`` file containing the trusted CAs, or ``False`` to disable all TLS verification
        :type verify: str|bool
        :param cache_ttl: Time in seconds to live for cached API responses
        :type cache_ttl: int
        :param certauth: The filepath for the client side certificate and the private key either as a single file or as a tuple of both files' paths
        :type certauth: str|tuple(str, str)
        :param custom_headers: A dictionary of any headers you want to send in **every** request
        :type custom_headers: dict
        :param kwargs: A dictionary of any other keyword arguments including;

            * **max_connection_retries** (*int*) - Number of attempts to retry when connecting to SOAR. Use ``-1`` for unlimited retries. Defaults to ``-1``.
            * **request_max_retries** (*int*) - Max number of times to retry a request to SOAR before exiting. Defaults to ``5``.
            * **request_retry_delay** (*int*) - Number of seconds to wait between retries. Defaults to ``2``.
            * **request_retry_backoff** (*int*) - Multiplier applied to delay between retry attempts. Defaults to ``2``.

        :type kwargs: dict
        """
        super(SimpleClient, self).__init__(org_name, base_url, proxies, verify, certauth, custom_headers=custom_headers, **kwargs)
        self.cache = TTLCache(maxsize=128, ttl=cache_ttl)

    def connect(self, email, password, timeout=None):
        """
        Connect and authenticate to the IBM SOAR REST API service.

        :param email: The email address to use for authentication.
        :type email: str
        :param password: The password.
        :type password: str
        :param timeout: optional timeout (seconds)
        :type timeout: int
        :return: The IBM SOAR session object.
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Calling connect of BaseClient. Convert the exception if there is any
        ret = None
        try:
            ret = super(SimpleClient, self).connect(email, password, timeout)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())

        return ret

    def __make_headers(self, co3_context_token=None, additional_headers=None):
        return self.make_headers(co3_context_token, additional_headers)

    def _keyfunc(self, uri, *args, **kwargs):
        """ function to generate cache key for cached_get """
        return uri

    def _get_cache(self):
        return self.cache

    def get(self, uri, co3_context_token=None, timeout=None, is_uri_absolute=None, get_response_object=None):
        """Gets the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to fetch.
        :type uri: str
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :param is_uri_absolute: if True, does not insert /org/{org_id} into the uri.
        :type is_uri_absolute: bool
        :param get_response_object: if True, returns entire response object.
        :type get_response_object: bool
        :return: A dictionary, list, or response object with the value returned by the server.
        :rtype: dict | list | Response
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call get from BaseClient, convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).get(uri, co3_context_token, timeout, is_uri_absolute, get_response_object)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    @cachedmethod(_get_cache, key=_keyfunc)
    def cached_get(self, uri, co3_context_token=None, timeout=None):
        """ Same as :meth:`get()`, but checks cache first """
        return self.get(uri, co3_context_token, timeout)

    def get_const(self, co3_context_token=None, timeout=None):
        """
        Get the ``ConstREST`` endpoint.

        Endpoint for retrieving various constant information for this server. This information is
        useful in translating names that the user sees to IDs that other REST API endpoints accept.

        For example, the ``incidentDTO`` has a field called ``"crimestatus_id"``. The valid values are stored
        in ``constDTO.crime_statuses``.

        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: ``ConstDTO`` as a dictionary
        :rtype: dict
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call get_const from BaseClient. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).get_const(co3_context_token, timeout)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    def get_content(self, uri, co3_context_token=None, timeout=None):
        """Gets the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to fetch.
        :type uri: str
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: The raw value returned by the server for this resource.
        :rtype: dict
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call get_content from BaseClient. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).get_content(uri, co3_context_token, timeout)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    def post(self, uri, payload, co3_context_token=None, timeout=None, headers=None):
        """
        Posts to the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to post.
        :type uri: str
        :param payload: A dictionary value to be posted.
        :type payload: dict
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :param headers: Optional headers to include.
        :type headers: dict
        :return: A dictionary or list with the value returned by the server.
        :rtype: dict | list
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call post of BaseClient. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).post(uri, payload, co3_context_token, timeout, headers=headers)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    def _patch(self, uri, patch, co3_context_token=None, timeout=None):
        """Internal method used to call the underlying server patch endpoint"""
        url = u"{0}/rest/orgs/{1}{2}".format(self.base_url, self.org_id, ensure_unicode(uri))
        if isinstance(patch, dict):
            payload_json = json.dumps(patch)
        else:
            payload_json = json.dumps(patch.to_dict())

        hdrs = {"handle_format": "names"}
        response = self._execute_request(self.session.patch,
                                         url,
                                         data=payload_json,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token,
                                                                   additional_headers=hdrs),
                                         verify=self.verify,
                                         timeout=timeout)

        return response

    def _handle_patch_response(self, response, patch, callback):
        """Helper to determine if a patch retry is needed.  Will only return True if the server responds with a
        409 or if the patch apply failed with field failures and the caller asked to overwrite conflicts."""
        if response.status_code == 409:
            # Just retry again, no adjustments.  The server can return 409 if there is a DB-level conflict.
            LOG.info("Retrying patch unchanged due to server CONFLICT")
            return True

        if response.status_code == 200:
            json = response.json()

            LOG.debug(json)

            patch_status = PatchStatus(json)

            if not patch_status.is_success() and patch_status.has_field_failures():
                LOG.info("Patch conflict detected - invoking callback")

                before = patch.get_old_values()

                try:
                    callback(response, patch_status, patch)
                except NoChange:
                    # Callback explicitly indicated that it didn't want to apply the change, so just
                    # return False here to stop processing.
                    #
                    LOG.debug("callback indicated no change after conflict - skipping")
                    return False

                if not patch.has_changes():
                    LOG.debug("callback removed all conflicts from patch - no need to re-issue")
                    return False

                # Make sure something in the patch has actually changed, otherwise we'd
                # just re-issue the same patch and get into a loop.
                after = patch.get_old_values()

                if before == after:
                    raise ValueError("invoked callback did not change the patch object, but returned True")

                return True

        # Raise an exception if there's some non-200 response.
        _raise_if_error(response)

        # Don't want to retry and got a 200 response.  There may or may not be field_failures.
        # Handling that is now up to the caller of the patch method.
        return False

    @staticmethod
    def _patch_overwrite_callback(response, patch_status, patch):
        """
        Callback to use when the caller specified overwrite_conflict=True in the patch call.
        """
        patch.update_for_overwrite(patch_status)

    @staticmethod
    def _patch_raise_callback(response, patch_status, patch):
        """
        Callback to use when the caller specified overwrite_conflict=False in the patch call.
        """
        # Got a conflict and no callback specified.  Just raise an exception.
        raise PatchConflictException(response, patch_status)

    def patch(self, uri, patch, co3_context_token=None, timeout=None, overwrite_conflict=False):
        """
        PATCH request to the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to patch.
        :type uri: str
        :param patch: The :class:`~resilient.patch.Patch` object to apply
        :type patch: :class:`~resilient.patch.Patch`
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :param overwrite_conflict: always overwrite fields in conflict. If ``True``, the passed-in patch
            object will be modified if necessary.
        :type overwrite_conflict: bool
        :return: the ``response`` from the endpoint.
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
        :raises SimpleHTTPException: if an HTTP exception or patch conflict occurs.
        :raises PatchConflictException: If the patch failed to apply (and overwrite_conflict is False).
        """
        if overwrite_conflict:
            # Re-issue patch with intent to overwrite conflicts.
            callback = SimpleClient._patch_overwrite_callback
        else:
            # Raise an exception on conflict.
            callback = SimpleClient._patch_raise_callback

        return self.patch_with_callback(uri, patch, callback, co3_context_token, timeout)

    def patch_with_callback(self, uri, patch, callback, co3_context_token=None, timeout=None):
        """
        PATCH request to the specified URI.  If the patch application fails because of field conflicts,
        the specified callback is invoked, allowing the caller to adjust the patch as necessary.

        :param uri: Relative URI of the resource to patch.
        :type str: Relative URI of the resource to patch.
        :param patch: The :class:`~resilient.patch.Patch` object to apply
        :type patch: :class:`~resilient.patch.Patch`
        :param callback: Function/lambda to invoke when a patch conflict is detected.  The function/lambda must be
          of the following form: ``def my_callback(response, patch_status, patch)``.
          If your callback raises :class:`resilient.NoChange <resilient.co3base.NoChange>`, the update is skipped.
        :type callback: func
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: the ``response`` from the endpoint.
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
        """
        response = self._patch(uri, patch, co3_context_token, timeout)

        while self._handle_patch_response(response, patch, callback):
            response = self._patch(uri, patch, co3_context_token, timeout)

        return response

    def post_attachment(self, uri, filepath, filename=None,
                        mimetype=None, data=None, co3_context_token=None, timeout=None,
                        bytes_handle=None):
        """
        Upload a file to the specified URI
        e.g. ``/incidents/<id>/attachments`` (for incident attachments)
        or ``/tasks/<id>/attachments`` (for task attachments)

        .. warning::
            Please see our updated :class:`resilient_lib.write_file_attachment <resilient_lib.components.resilient_common.write_file_attachment>`
            as this method takes a data stream and is more reliable

        :param uri: Relative URI of the resource to post.
        :type uri: str
        :param filepath: the path of the file to post
        :type filepath: str
        :param filename: optional name of the file when posted
        :type filename: str
        :param mimetype: optional override for the guessed MIME type
        :type mimetype: str
        :param data: optional dict with additional ``MIME`` parts (not required for file attachments; used in artifacts)
        :type data: dict
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :param bytes_handle: BytesIO handle for content or use filepath
        :type bytes_handle: BytesIO
        :return: the ``response`` from the endpoint.
        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
        """
        # Call BaseClient post_attachment. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).post_attachment(uri, filepath, filename=filename,
                                                                 mimetype=mimetype,
                                                                 data=data,
                                                                 co3_context_token=co3_context_token,
                                                                 timeout=timeout,
                                                                 bytes_handle=bytes_handle)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    def search(self, payload, co3_context_token=None, timeout=None):
        """
        Posts to the ``SearchExREST`` endpoint.

        Endpoint for performing full text searches through incidents and incident child objects
        (tasks, incident comments, task comments, milestones, artifacts, incident attachments,
        task attachments, and data tables).

        :param payload: The SearchExInputDTO parameters for performing a search.
        :type payload: dict
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: List of results, as an array of ``SearchExResultDTO``
        :rtype: list
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        url = u"{0}/rest/search_ex".format(self.base_url)
        payload_json = json.dumps(payload)
        response = self._execute_request(self.session.post,
                                         url,
                                         data=payload_json,
                                         proxies=self.proxies,
                                         cookies=self.cookies,
                                         headers=self.make_headers(co3_context_token),
                                         verify=self.verify,
                                         timeout=timeout)
        _raise_if_error(response)
        return json.loads(response.text)

    def get_put(self, uri, apply_func, co3_context_token=None, timeout=None):
        """
        Safely performs an update operation by a GET, calls your ``apply_func`` callback, then PUT
        with the updated value.  If the put call returns a ``409`` error, these steps are retried.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to get and update.
        :type uri: str
        :param apply_func: A callback function that you implement to update the resource.  The function must be
          of the following form: ``def my_apply_func(object_to_update)``, and update the object.
          If your callback raises :class:`resilient.NoChange <resilient.co3base.NoChange>`, the update is skipped.
        :type apply_func: func
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :return: A dictionary or list with the value returned by the PUT operation.
        :rtype: dict | list
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call BaseClient get_put. Convert exception if there is any
        res = None
        try:
            res = super(SimpleClient, self).get_put(uri, apply_func, co3_context_token, timeout)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return res

    def put(self, uri, payload, co3_context_token=None, timeout=None, headers=None):
        """
        Directly performs an update operation by PUT to the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``


        :param uri: Relative URI of the resource to update.
        :type uri: str
        :param payload: The object to update.
        :type payload: dict
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :param headers: Optional headers to include.
        :type headers: dict
        :return: A dictionary or list with the value returned by the server.
        :rtype: dict | list
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call BaseClient put. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).put(uri, payload, co3_context_token, timeout, headers=headers)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response

    def delete(self, uri, co3_context_token=None, timeout=None):
        """
        Deletes the specified URI.

        .. note::
            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
            if you specify a uri of ``/incidents``, the actual URL would be something like:
            ``https://soar.ibm.com/rest/orgs/201/incidents``

        :param uri: Relative URI of the resource to delete.
        :type uri: str
        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
        :type co3_context_token: str
        :param timeout: Optional timeout (seconds).
        :type timeout: int
        :raises SimpleHTTPException: if an HTTP exception occurs.
        """
        # Call BaseClient delete. Convert exception if there is any
        response = None
        try:
            response = super(SimpleClient, self).delete(uri, co3_context_token, timeout)
        except co3base.BasicHTTPException as ex:
            _raise_if_error(ex.get_response())
        return response


class LoggingSimpleClient(SimpleClient):
    """ Simple Client version that logs all Resilient REST API responses to disk.  Useful when building a Mock."""
    def __init__(self, logging_directory="", *args, **kwargs):
        super(LoggingSimpleClient, self).__init__(*args, **kwargs)
        try:
            directory = os.path.expanduser(logging_directory)
            directory = os.path.expandvars(directory)
            assert(os.path.exists(directory))
            self.logging_directory = directory
        except Exception as e:
            raise Exception("Response Logging Directory %s does not exist!",
                            logging_directory)

    def _log_response(self, response, *args, **kwargs):
        """ Log Headers and JSON from a Requests Response object """
        url = urlparse.urlparse(response.url)
        filename = "_".join((str(response.status_code), "{0}",
                             response.request.method,
                             url.path, url.params,
                             datetime.datetime.now().isoformat())).replace('/', '_').replace(':', '-')
        try:
            with open(os.path.join(self.logging_directory,
                                   filename.format("JSON")), "w+") as logfile:
                logfile.write(json.dumps(response.json(), indent=2))
        except:
            with open(os.path.join(self.logging_directory,
                                   filename.format("DATA")), "w+b") as logfile:
                logfile.write(response.content)
        with open(os.path.join(self.logging_directory,
                               filename.format("HEADER")), "w+") as logfile:
            logfile.write(json.dumps(dict(response.headers), indent=2))

    def _connect(self, *args, **kwargs):
        """ Connect to Resilient and log response """
        normal_post = self.session.post
        self.session.post = lambda *args, **kwargs: normal_post(
            hooks=dict(response=self._log_response), *args, **kwargs)
        session = super(LoggingSimpleClient, self)._connect(*args, **kwargs)
        self.session.post = normal_post
        return session

    def _execute_request(self, operation, url, **kwargs):
        """Execute a HTTP request and log response.
           If unauthorized (likely due to a session timeout), retry.
        """
        def wrapped_operation(url, **kwargs):
            return operation(url, hooks=dict(response=self._log_response),
                             **kwargs)
        return super(LoggingSimpleClient, self)._execute_request(
            wrapped_operation, url, **kwargs)



#!/usr/bin/env python
# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

import warnings
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = None

from resilient import *

warnings.warn("The 'co3' module is deprecated, use 'resilient' instead", DeprecationWarning)



# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.
from __future__ import print_function

import doctest
import os
import time
import types
from argparse import ArgumentParser
from shutil import rmtree

import pytest
import resilient
from mock import patch

DEFAULT_CONFIG_FILENAME = "app.config"
DEFAULT_CONFIG_FILE = os.path.expanduser(os.path.join("~", ".resilient", DEFAULT_CONFIG_FILENAME))

def mock_os_path_exists(path):
    return True

def mock_os_path_not_exists(path):
    return False

def mock_makedirs(dir):
    pass

class xTestCo3:
    """Basic API tests"""

    def test_doctest(self):
        """Run doctest on all modules, even if you didn't --doctest-modules"""
        for item in resilient.__dict__.values():
            if type(item) is types.ModuleType:
                doctest.testmod(item, verbose=True, raise_on_error=True)

    def test_connect_no_verify(self, co3_args):
        """ Successful connection with no Cert Verification """
        url = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=url,
                                        verify=False)
        assert client
        userinfo = client.connect(co3_args.email, co3_args.password)
        assert userinfo

    def test_connect_proxy(self, co3_args):
        """ Successful connection with no Cert Verification """
        url = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=url,
                                        verify=False,
                                        proxies=co3_args.proxy)
        assert client
        user_info = client.connect(co3_args.email, co3_args.password)
        assert user_info

    @pytest.mark.parametrize("resp", ([], None))
    def test_extract_org_id_failure(self, co3_args, resp):
        """Extract org id given org name"""
        url = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=url,
                                        verify=False,
                                        proxies=co3_args.proxy)
        try:
            client._extract_org_id({"orgs": resp})
            assert False
        except Exception as e:
            assert True

    def test_org_error(self, co3_args):
        """No org entered"""
        url = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=url,
                                        verify=False,
                                        proxies=co3_args.proxy)
        # Mismatch.
        try:
            client._extract_org_id({"orgs": [{"name": "Fake Org"}]})
            assert False
        except Exception as e:
            assert True

        # not enabled
        try:
            client._extract_org_id({"orgs": [{"name": client.org_name,
                                              "enabled": False}]})
            assert False
        except Exception as e:
            assert True

        # Erase org_name input
        client.org_name = None
        try:
            client._extract_org_id({"orgs": [{"name": "Fake Org"}]})
            assert False
        except Exception as e:
            assert True

    def test_api_key(self, co3_args):
        """
        Test using api key to get org
        :return:
        """
        print(str(co3_args))
        url = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=url,
                                        verify=False,
                                        proxies=co3_args.proxy)
        if (co3_args.api_key_id is not None and co3_args.api_key_secret is not None):
            try:
                # Erase org_id
                client.org_id = None
                ret = client.set_api_key(co3_args.api_key_id, co3_args.api_key_secret)
                assert client.org_id
            except Exception as e:
                assert False


class xTestCo3Patch:
    """Tests for patch, create_patch, and get_patch methods"""
    def _connect(self, co3_args):
        uri = "https://{0}:{1}".format(co3_args.host, co3_args.port or 443)
        client = resilient.SimpleClient(org_name=co3_args.org,
                                        base_url=uri,
                                        verify=False)
        assert client
        user_info = client.connect(co3_args.email, co3_args.password)
        assert user_info
        return client

    def _create_incident(self, client, incident_dict):
        """create a test incident to use for patching"""
        incident = {"name": __name__,
                    "discovered_date": int(time.time() * 1000)
        }
        incident.update(incident_dict)
        inc = client.post("/incidents", incident)
        return inc

    @pytest.mark.parametrize("overwrite_conflict", (True, False))
    def test_patch_conflict(self, co3_args, overwrite_conflict):
        """ do incident patch that results in conflict """
        client = self._connect(co3_args)
        inc = self._create_incident(client, {"name": "test"})
        uri = "/incidents/%d" % inc['id']
        # Create a conflict
        inc["name"] = "the wrong value"
        inc["vers"] -= 1 # Force it to check old_value
        patch = resilient.Patch(inc)
        patch.add_value("name", "test updated")

        if overwrite_conflict:
            # If overwrite_conflict is specified then patch will return.
            response = client.patch(uri, patch, overwrite_conflict=overwrite_conflict)

            assert resilient.PatchStatus(response.json()).is_success() is overwrite_conflict
        else:
            # Not overwriting conflict, so an exception will be thrown.
            with pytest.raises(resilient.PatchConflictException) as exception_info:
                client.patch(uri, patch, overwrite_conflict=overwrite_conflict)

            # Gather the patch_status value from the exception for additional verification.
            patch_status = exception_info.value.patch_status

            fail_msg = "could not be applied due to a conflicting edit by another user.  The following field(s) were in conflict:  name."
            assert fail_msg in patch_status.get_message()

            assert patch_status.get_conflict_fields() == ["name"]
            assert patch_status.get_your_original_value("name") == "the wrong value"
            assert patch_status.get_actual_current_value("name") == "test"

        inc = client.get("/incidents/%d" % inc['id'])

        if overwrite_conflict:
            assert inc['name'] == "test updated"
        else:
            assert inc['name'] == "test"

    def test_patch_no_conflict(self, co3_args):
        """ do incident_patch with no conflict """
        client = self._connect(co3_args)
        inc = self._create_incident(client, {"name": "test"})
        uri = "/incidents/%d" % inc['id']
        patch = resilient.Patch(inc)
        patch.add_value("name", "test updated")
        response = client.patch(uri, patch, overwrite_conflict=False)
        assert resilient.PatchStatus(response.json()).is_success()
        inc = client.get("/incidents/%d" % inc['id'])
        assert inc['name'] == "test updated"

    def test_patch_null_old_value(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test", "description": None})

        patch = resilient.Patch(inc)

        patch.add_value("description", "new value")

        uri = "/incidents/%d" % inc['id']

        response = client.patch(uri, patch)

        inc = client.get("/incidents/%d" % inc['id'])

        assert inc["description"] == "new value"

    def test_patch_invalid_callback(self, co3_args):
        """
        If a callback returns True but didn't modify the passed in patch in any way, that'd be a problem.
        So make sure we throw an exception in that case.
        """
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test"})

        uri = "/incidents/%d" % inc['id']

        # Create a conflict
        inc["name"] = "the wrong value"
        inc["vers"] -= 1 # Force it to check old_value

        patch = resilient.Patch(inc)

        patch.add_value("name", "test updated")

        def mycb(response, patch_status, patch):
            # Return True but don't modify the patch.
            return True

        with pytest.raises(ValueError) as exception_info:
            client.patch_with_callback(uri, patch, mycb)

        assert "invoked callback did not change the patch object, but returned True" in str(exception_info.value)

    def test_no_change(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test"})

        uri = "/incidents/%d" % inc['id']

        # Create a conflict
        inc["name"] = "the wrong value"
        inc["vers"] -= 1 # Force it to check old_value

        patch = resilient.Patch(inc)

        patch.add_value("name", "test updated")

        def mycb(response, patch_status, patch):
            raise resilient.NoChange

        response = client.patch_with_callback(uri, patch, mycb)

        assert response
        assert response.status_code == 200

        patch_status = resilient.PatchStatus(response.json())

        assert not patch_status.is_success()
        assert patch_status.get_conflict_fields() == ["name"]

    def test_conflict_with_handler(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test"})

        uri = "/incidents/%d" % inc['id']

        # Create a conflict
        inc["name"] = "the wrong value"
        inc["vers"] -= 1 # Force it to check old_value

        patch = resilient.Patch(inc)

        patch.add_value("name", "test updated")

        def mycb(response, patch_status, patch):
            patch.exchange_conflicting_value(patch_status, "name", "test updated take 2")

        response = client.patch_with_callback(uri, patch, mycb)

        assert response
        assert response.status_code == 200

        assert "test updated take 2" == client.get(uri)["name"]

    def test_delete(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test"})

        uri = "/incidents/%d" % inc['id']

        response = client.delete(uri)
        assert response
        assert response['success'] == True

    def test_post_attachment(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test for attachment"})

        file_name = "test-for-attachment.txt"
        file_content = "this is test data"

        # Create the file
        temp_file = open(file_name, "w")
        temp_file.write(file_content)
        temp_file.close()
        # Post file to Resilient
        response = client.post_attachment("/incidents/{0}/attachments".format(inc["id"]),
                                          file_name,
                                          file_name,
                                          mimetype="text/plain")

        assert response
        assert response['name'] == file_name

        os.remove(file_name)

    def test_post_attachment_artifact(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "test for artifact attachment"})

        file_name = "test-for-attachment_artifact.txt"
        file_content = "this is test data"

        # Create the file
        temp_file = open(file_name, "w")
        temp_file.write(file_content)
        temp_file.close()
        # Post file to Resilient
        attachment_uri = "/incidents/{0}/artifacts/files".format(inc["id"])

        response = client.post_artifact_file(attachment_uri,
                                             artifact_type=16,
                                             artifact_filepath=file_name,
                                             description="Description",
                                             value=file_name)

        assert response

        os.remove(file_name)

    def test_get_config_file(self, co3_args):
        config_file = resilient.get_config_file("~/.resilient/app.config")

        assert config_file

    def test_get_client(self, co3_args):
        client = resilient.get_client(co3_args)

        assert client

    def test_throw_simple_http_exception(self, co3_args):
        with pytest.raises(resilient.SimpleHTTPException) as exception_info:
            client = resilient.SimpleClient("Not A real Org")
            user_info = client.connect("not_a_user@mail.com", "test")

            assert exception_info

    def test_get_const(self, co3_args):
        client = self._connect(co3_args)

        response = client.get_const()

        assert response

    def test_get_content(self, co3_args):
        client = self._connect(co3_args)

        response = client.get_content("/incidents")

        assert response

    def test_get_proxy_dict(self, co3_args):
        proxy_opts = type('', (), {})()
        proxy_opts.proxy_host = "http://resilientproxy.com"
        proxy_opts.proxy_port = "4443"
        proxy_opts.proxy_user = None
        proxy_opts.proxy_password = None
        proxy_dict = resilient.get_proxy_dict(proxy_opts)

        assert proxy_dict == {'https': 'http://resilientproxy.com:4443'}

        proxy_opts.proxy_user = "user"
        proxy_opts.proxy_password = "password"
        proxy_dict = resilient.get_proxy_dict(proxy_opts)

        assert proxy_dict == {'https': 'http://user:password@resilientproxy.com:4443/'}

        proxy_opts.proxy_host = "resilientproxy.com"

        proxy_dict = resilient.get_proxy_dict(proxy_opts)

        assert proxy_dict == {'https': 'https://user:password@resilientproxy.com:4443/'}

    def test_changing_description(self, co3_args):
        client = self._connect(co3_args)

        inc = self._create_incident(client, {"name": "Test Put and get_put"})
        # use PUT to change description
        inc["description"] = "Test Put method"
        uri = "/incidents/{}".format(inc["id"])
        inc = client.put(uri, inc)

        # use get_put to change description

        def change_description(json_data):
            json_data["description"] = json_data.get("description") + ", test get_put method"

        inc = client.get_put(uri, change_description)

        # Test no change error

        def no_change(json_data):
            raise resilient.NoChange

        inc = client.get_put(uri, no_change)

    @patch('resilient.SimpleClient.connect')
    @patch('resilient.SimpleClient.get')
    def test_get_client_dict(self, mock_get, mock_connect, co3_args):
        mock_connect.return_value = {"orgs": [{"id": 204}]}
        mock_get.return_value = {"actions_framework_enabled": True}

        args = {
            'host': 'resilient_host',
            'email': 'a@example.com',
            'password': 'password',
            'org': 'org',
        }

        rest_client = resilient.get_client(args)
        assert not rest_client.proxies

        #
        args['proxy_host'] = 'proxy_host'
        args['proxy_port'] = 1443

        rest_client = resilient.get_client(args)
        assert rest_client.proxies
        assert rest_client.proxies['https'] == "https://proxy_host:1443"

        #
        args['proxy_user'] = 'proxy_user'
        args['proxy_password'] = 'proxy_password'

        rest_client = resilient.get_client(args)
        assert rest_client.proxies
        assert rest_client.proxies['https'] == "https://proxy_user:proxy_password@proxy_host:1443/"

    @patch('resilient.SimpleClient.connect')
    @patch('resilient.SimpleClient.get')
    def test_get_client_namespace(self, mock_get, mock_connect, co3_args):
        mock_connect.return_value = {"orgs": [{"id": 204}]}
        mock_get.return_value = {"actions_framework_enabled": True}

        parser = ArgumentParser()
        parser.add_argument('--proxy_host')
        parser.add_argument('--proxy_port')
        parser.add_argument('--proxy_user')
        parser.add_argument('--proxy_password')
        parser.add_argument('--email')
        parser.add_argument('--password')
        parser.add_argument('--host')
        parser.add_argument('--org')

        #
        args = parser.parse_args([
            '--email', 'a@example.com',
            '--password', 'password',
            '--host', 'resilient_host',
            '--org', 'org'
        ])

        rest_client = resilient.get_client(args)
        assert not rest_client.proxies

        #
        args = parser.parse_args([
            '--email', 'a@example.com',
            '--password', 'password',
            '--host', 'resilient_host',
            '--org', 'org',
            '--proxy_host', 'proxy_host',
            '--proxy_port', '1443'
        ])

        rest_client = resilient.get_client(args)
        assert rest_client.proxies
        assert rest_client.proxies['https'] == "https://proxy_host:1443"

        #
        args = parser.parse_args([
            '--proxy_host', 'proxy_host',
            '--proxy_port', '1443',
            '--proxy_user', 'proxy_user',
            '--proxy_password', 'proxy_password',
            '--email', 'a@example.com',
            '--password', 'password',
            '--host', 'resilient_host',
            '--org', 'org'
        ])

        rest_client = resilient.get_client(args)
        assert rest_client.proxies
        assert rest_client.proxies['https'] == "https://proxy_user:proxy_password@proxy_host:1443/"


class TestGetConfig(object):

    @pytest.fixture(scope='session')
    def tmp_config_file(self, tmpdir_factory):
        tmp_config_dir = tmpdir_factory.mktemp(".resilient")
        tmp_config_file = os.path.expanduser(os.path.join(tmp_config_dir, DEFAULT_CONFIG_FILENAME))
        yield tmp_config_file
        rmtree(str(tmp_config_dir))

    def xtest_with_env_variable(self, monkeypatch, tmp_config_file):
        """
        Test with environment variable $APP_CONFIG_FILE.
        """
        monkeypatch.setenv("APP_CONFIG_FILE", str(tmp_config_file))
        result = resilient.get_config_file()
        assert result == tmp_config_file

    @patch('resilient.co3.os.path.exists', side_effect=mock_os_path_exists)
    def test_with_filename_exists_local(self, mock_os_path_exists):
        """
        Test with filename specified and not exists in home path.
        """
        test_file_name = "test_file"
        result = resilient.get_config_file(filename=test_file_name)
        assert result == test_file_name

    @patch('resilient.co3.os.path.exists', side_effect=mock_os_path_not_exists)
    def test_with_filename_not_exists_local(self, mock_os_path_not_exists):
        """
        Test with filename specified and exists in home path.
        """
        test_file_name = "test_file"
        test_file_abs = os.path.expanduser(os.path.join("~", ".resilient", test_file_name))
        result = resilient.get_config_file(filename=test_file_name)
        assert result == test_file_abs

    @patch('resilient.co3.os.path.exists', side_effect=mock_os_path_exists)
    def test_with_app_config_exists_local(self, mock_os_path_exists):
        """
        Test with default config specified and exists in home path.
        """
        result = resilient.get_config_file()
        assert result == DEFAULT_CONFIG_FILENAME

    @patch('resilient.co3.os.path.exists', side_effect=mock_os_path_not_exists)
    def test_with_app_config_not_exists_local(self, mock_os_path_not_exists):
        """
        Test with default config specified and not exists in home path.
        """
        result = resilient.get_config_file()
        assert result == DEFAULT_CONFIG_FILE



# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.
import pytest
import unittest
import mock
from resilient.co3sslutil import match_hostname


class TestCo3SSLUtil(unittest.TestCase):
    def test_match_hostname(self):
        with pytest.raises(Exception):
            match_hostname(mock.Mock(), "test.com")



import json
from mock import patch

import resilient


@patch("resilient.co3base.BaseClient.set_api_key")
@patch("resilient.co3base.BaseClient.get")
def test_get_client(mock_get_call, mock_session):

    mock_cert = ("mock_cert_file_path.pem", "mock_cert_file_key_path.pem")
    mock_opts = {
        "org": "Test Organization",
        "host": "example.com",
        "port": 8080,
        "api_key_id": "1234-efgh",
        "api_key_secret": "abcd-56789",
        "cafile": "FALSE",
        "client_auth_cert": mock_cert[0],
        "client_auth_key": mock_cert[1]
    }
    mock_session_info = {
        "orgs": [{"name": "Test Organization", "id": 201}]
    }
    mock_get_org_data = {
        "actions_framework_enabled": False
    }
    mock_session.return_value = mock_session_info
    mock_get_call.return_value = mock_get_org_data

    res_client = resilient.get_client(mock_opts)

    assert res_client.base_url == "https://example.com:8080"
    assert res_client.cert == mock_cert
    assert res_client.actions_enabled == False

def test_client_put(fx_simple_client):

    base_client = fx_simple_client[0]
    requests_adapter = fx_simple_client[1]

    test_headers = {"content-type": "application/octet-stream"}

    mock_uri = '{0}/rest/orgs/{1}/playbooks/imports'.format(base_client.base_url, base_client.org_id)    

    requests_adapter.register_uri('PUT',
                                  mock_uri,
                                  status_code=200,
                                  request_headers=test_headers,
                                  text=json.dumps(test_headers))

    uri = "/playbooks/imports"

    r = base_client.put(
        uri,
        "payload",
        headers=test_headers
    )

    assert r['content-type'] == test_headers['content-type']

def test_client_post(fx_simple_client):
    base_client = fx_simple_client[0]
    requests_adapter = fx_simple_client[1]
    incident_id = 1001

    mock_uri = '{0}/rest/orgs/{1}/incidents/{2}'.format(base_client.base_url, base_client.org_id, incident_id)
    mock_response = {"incident_id": incident_id}
    requests_adapter.register_uri('POST', mock_uri, status_code=200, text=json.dumps(mock_response))
    r = base_client.post("/incidents/1001", {"incident_name": "Mock Incident"})

    assert r.get("incident_id") == 1001

def test_simple_client_get_const(fx_simple_client):
    base_client = fx_simple_client[0]
    requests_adapter = fx_simple_client[1]

    mock_uri = '{0}/rest/const'.format(base_client.base_url)
    mock_response = {"server_version": {"major": 47, "minor": 0, "build_number": 8304, "version": "47.0.8304"}}
    requests_adapter.register_uri('GET', mock_uri, status_code=200, text=json.dumps(mock_response))
    r = base_client.get_const()

    assert r.get("server_version", {}).get("major") == 47



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

import os
import pkg_resources

PACKAGE_NAME = "resilient"

ENV_HTTP_PROXY = "HTTP_PROXY"
ENV_HTTPS_PROXY = "HTTPS_PROXY"
ENV_NO_PROXY = "NO_PROXY"

PROTECTED_SECRET_PREFIX = "$"
ALLOW_UNRECOGNIZED = False
MIN_SUPPORTED_PY3_VERSION = (3, 6)

# ENV Vars
ENV_VAR_APP_HOST_CONTAINER = "APP_HOST_CONTAINER"

# Headers
HEADER_MODULE_VER_KEY = "Resilient-Module-Version"
HEADER_MODULE_VER_VALUE = pkg_resources.get_distribution(PACKAGE_NAME).version

# Error Codes
ERROR_CODE_CONNECTION_UNAUTHORIZED = 21

# Error Messages
ERROR_MSG_CONNECTION_UNAUTHORIZED = u"Unauthorized"
ERROR_MSG_CONNECTION_INVALID_CREDS = u"Either the API Key has been blocked, the API Credentials are incorrect or the IP address has been banned. Please review the SOAR logs for more information"
WARNING_PROTECTED_SECRETS_NOT_SUPPORTED = u"Protected secrets are only Python >= 3 supported."

# File Paths
PATH_SECRETS_DIR = os.path.join(os.path.abspath(os.sep), "etc", "secrets")
PATH_JWK_FILE = os.path.join(PATH_SECRETS_DIR, ".jwk", "key.jwk")

# app configs keys
APP_CONFIG_MAX_CONNECTION_RETRIES = "max_connection_retries"
APP_CONFIG_REQUEST_MAX_RETRIES = "request_max_retries"
APP_CONFIG_REQUEST_RETRY_DELAY = "request_retry_delay"
APP_CONFIG_REQUEST_RETRY_BACKOFF = "request_retry_backoff"

# app config default values
APP_CONFIG_MAX_CONNECTION_RETRIES_DEFAULT = -1
APP_CONFIG_REQUEST_MAX_RETRIES_DEFAULT = 5
APP_CONFIG_REQUEST_RETRY_DELAY_DEFAULT = 2
APP_CONFIG_REQUEST_RETRY_BACKOFF_DEFAULT = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""
Used by resilient-circuits for customize.
Used by resilient-sdk for reloading the customize.py file
"""


class Definition(object):
    """A definition that can be loaded for resilient-circuits customize."""

    def __init__(self, value):
        self.value = value


class TypeDefinition(Definition):
    """Definition of a type with fields"""
    pass


class MessageDestinationDefinition(Definition):
    """Definition of a message destination"""
    pass


class FunctionDefinition(Definition):
    """Definition of a function"""
    pass


class ActionDefinition(Definition):
    """Definition of a custom action (rule)"""
    pass


class ImportDefinition(Definition):
    """Definition of a set of importable customizations"""
    pass



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.

"""Command-line argument parser for Resilient apps"""

import argparse
import getpass
import logging
import os
import sys

import keyring
from six import string_types

from resilient import constants, helpers

if sys.version_info.major == 2:
    from io import open

    from resilient import (ensure_unicode, get_proxy_dict,
                           get_resilient_circuits_version)
else:
    from resilient.co3 import (ensure_unicode, get_proxy_dict,
                               get_resilient_circuits_version)

try:
    # For all python < 3.2
    import backports.configparser as configparser
except ImportError:
    import configparser

logger = logging.getLogger(__name__)


class ConfigDict(dict):
    """A dictionary, with property-based accessor

    >>> opts = {"one": 1}
    >>> cd = ConfigDict(opts)
    >>> cd["one"]
    1
    >>> cd.one
    1

    """
    def __getattr__(self, name):
        """Attributes are made accessible as properties"""
        try:
            return self[name]
        except KeyError:
            raise AttributeError()


class ArgumentParser(argparse.ArgumentParser):
    """Helper to parse common Resilient command line arguments.

    Optionally, arguments can be specified in a config file, section :samp:`[resilient]`.

    It is expected that a command line utility that needs to work with a Resilient
    server will create its own class that inherits this class.  Its `__init__`
    method should call :samp:`self.add_argument(...)` as necessary.

    See https://docs.python.org/3/library/argparse.html for more details.
    """

    DEFAULT_PORT = 443
    config = None

    def getopt(self, section, opt):
        """Get a single option value.

        :param section: The configuration section.
        :param opt: The configuration option.
        :return: The value, or None if not present.
        """
        if self.config:
            if opt in self.config.options(section):
                return self.config.get(section, opt)
        return None

    def getopts(self, section, opt):
        """Get a comma-separated option value as an array.

        :param section: The configuration section.
        :param opt: The configuration option.
        :return: The value, or [] if not present.
        """
        if self.config:
            if opt in self.config.options(section):
                return self.config.get(section, opt).split(u",")
        return []

    def __init__(self, config_file=None):
        super(ArgumentParser, self).__init__()

        # Read configuration options.
        if config_file:
            config_path = ensure_unicode(config_file)
            config_path = os.path.expanduser(config_path)
            if os.path.exists(config_path):
                try:
                    self.config = configparser.ConfigParser(interpolation=None)
                    with open(config_path, 'r', encoding='utf-8') as f:
                        first_byte = f.read(1)
                        if first_byte != u'\ufeff':
                            # Not a BOM, no need to skip first byte
                            f.seek(0)
                        self.config.read_file(f)
                except Exception as exc:
                    logger.warn(u"Couldn't read config file '%s': %s", config_path, exc)
                    self.config = None
            else:
                logger.warn(u"Couldn't read config file '%s'", config_file)

        default_email = self.getopt("resilient", "email")
        default_password = self.getopt("resilient", "password")
        default_key_id = self.getopt("resilient", "api_key_id")
        default_key_secret = self.getopt("resilient", "api_key_secret")
        default_host = self.getopt("resilient", "host")
        default_port = self.getopt("resilient", "port") or self.DEFAULT_PORT
        default_proxy = self.getopts("resilient", "proxy")
        default_org = self.getopt("resilient", "org")
        default_cafile = self.getopt("resilient", "cafile")
        default_cache_ttl = int(self.getopt("resilient", "cache_ttl") or 0)
        default_proxy_host = self.getopt("resilient", "proxy_host")
        default_proxy_port = self.getopt("resilient", "proxy_port") or 0
        default_proxy_user = self.getopt("resilient", "proxy_user")
        default_proxy_password = self.getopt("resilient", "proxy_password")
        default_stomp_prefetch_limit = int(self.getopt("resilient", "stomp_prefetch_limit") or 20)
        default_resilient_mock = self.getopt("resilient", "resilient_mock")

        default_max_request_retries = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_MAX_RETRIES) or constants.APP_CONFIG_REQUEST_MAX_RETRIES_DEFAULT
        default_request_retry_delay = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_RETRY_DELAY) or constants.APP_CONFIG_REQUEST_RETRY_DELAY_DEFAULT
        default_request_retry_backoff = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_RETRY_BACKOFF) or constants.APP_CONFIG_REQUEST_RETRY_BACKOFF_DEFAULT

        self.add_argument("--email",
                          default=default_email,
                          help="The email address to use to authenticate to the Resilient server.")

        self.add_argument("--password",
                          default=default_password,
                          help="WARNING:  This is an insecure option since the password "
                               "will be visible to other processes and stored in the "
                               "command history.  The password to use to authenticate "
                               "to the Resilient server.  If omitted, the you will be prompted.")

        self.add_argument("--api_key_id",
                          default=default_key_id,
                          help="The api key id for API key.")

        self.add_argument("--api_key_secret",
                          default=default_key_secret,
                          help="WARNING:  This is an insecure option since the key secret "
                               "will be visible to other processes and stored in the "
                               "command history.  The password to use to authenticate "
                               "to the Resilient server.  If omitted, the you will be prompted.")

        self.add_argument("--host",
                          default=default_host,
                          required=default_host is None,
                          help="Resilient server host name.")

        self.add_argument("--port",
                          type=int,
                          default=default_port,
                          help="Resilient server REST API port number.")

        self.add_argument("--proxy",
                          default=default_proxy,
                          nargs="*",
                          help="An optional HTTP proxy to use when connecting.")

        self.add_argument("--org",
                          default=default_org,
                          help="The name of the organization to use.  If you are a member "
                               "of just one organization, then you can omit this argument.")

        self.add_argument("--cafile",
                          default=default_cafile,
                          help="The name of a file that contains trusted certificates.")

        self.add_argument("--cache-ttl",
                          default=default_cache_ttl or 240,
                          type=int,
                          help="TTL for API responses when using cached_get")

        self.add_argument("--proxy_host",
                          default=default_proxy_host,
                          help="HTTP Proxy host for Resilient Connection.")

        self.add_argument("--proxy_port",
                          type=int,
                          default=default_proxy_port,
                          help="HTTP Proxy port for Resilient Connection.")

        self.add_argument("--proxy_user",
                          default=default_proxy_user,
                          help="HTTP Proxy username for Resilient connection authentication.")

        self.add_argument("--proxy_password",
                          default=default_proxy_password,
                          help="HTTP Proxy password for Resilient connection authentication.")

        self.add_argument("--stomp-prefetch-limit",
                          default=default_stomp_prefetch_limit,
                          type=int,
                          help="MAX number of Action Module messages to send before ACK is required")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_MAX_RETRIES),
                          type=int,
                          default=default_max_request_retries,
                          help="Max number of times to retry a request to SOAR before exiting. Defaults to 5")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_RETRY_DELAY),
                          type=int,
                          default=default_request_retry_delay,
                          help="Number of seconds to wait between retries. Defaults to 2")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_RETRY_BACKOFF),
                          type=int,
                          default=default_request_retry_backoff,
                          help="Multiplier applied to delay between retry attempts. Defaults to 2")

        v_resc = get_resilient_circuits_version()

        # Having --resilient-mock here allows us to run unit tests for resilient-circuits and resilient-sdk
        # This argument will be added if:
        # - resilient-circuits is not installed
        # - resilient-circuits is installed and is > 34
        if not v_resc or (v_resc and v_resc.get("major") > 34):
            self.add_argument("--resilient-mock",
                            default=default_resilient_mock, 
                            help="<path_to_mock_module>.NameOfMockClass")

    def parse_args(self, args=None, namespace=None, ALLOW_UNRECOGNIZED=False):
        """
        Parse the configuration options and command-line arguments.

        :return: Note: the return value is a dict, not a Namespace.
        """

        if ALLOW_UNRECOGNIZED or constants.ALLOW_UNRECOGNIZED:
            return self.parse_known_args(args, namespace)[0]

        # (the implementation calls parse_known_args)
        args = super(ArgumentParser, self).parse_args(args, namespace)
        return args

    def parse_known_args(self, args=None, namespace=None):
        """
        Parse the configuration options and command-line arguments.

        :return: Note: the return value is a dict, not a Namespace.
        """
        # Parse the known arguments
        args, argv = super(ArgumentParser, self).parse_known_args(args, namespace)
        return _post_process_args(args), argv


def _post_process_args(args):
    # Post-process any options that reference keyring or environment variables
    opts = parse_parameters(vars(args))
    args = ConfigDict(opts)

    # Post-processing for other special options
    password = args.password
    email = args.email

    while (not password and email) and (not args.get("no_prompt_password")):
        password = getpass.getpass()
    args["password"] = ensure_unicode(password)

    if args.get("cafile"):
        args["cafile"] = os.path.expanduser(args.cafile)

    if args.get("stomp_cafile"):
        args["stomp_cafile"] = os.path.expanduser(args.stomp_cafile)

    if args.get("proxy_host"):
        args["proxy"] = get_proxy_dict(args)

    return args


def parse_parameters(options):
    """Given a dict that has configuration keys mapped to values,
       - If a value begins with '^', redirect to fetch the value from
         the secret key stored in the keyring.
         The keyring service name is always just an underscore
         (so keys must be unique in the whole options dict)
       - If a value begins with '$', fetch the value from environment.

    >>> opts = {
    ...    "thing": u"value",
    ...    "key3": "^val3",
    ...    "key4": u"$val4",
    ...    "key5": "$val5",
    ...    "deep1": {"key1": "val1", "key2": u"^val2"}
    ... }

    >>> keyring.set_password("_", "val3", "key3password")
    >>> keyring.set_password("_", "val2", "")
    >>> keyring.set_password("deep1", "val2", "key2password")
    >>> os.environ["val4"] = "key4param"
    >>> os.environ["val5"] = "key5param"

    >>> str(parse_parameters(opts)["key3"])
    'key3password'

    >>> parse_parameters(opts)["deep1"]["key1"]
    'val1'

    >>> str(parse_parameters(opts)["deep1"]["key2"])
    'key2password'

    >>> parse_parameters(opts)["deep1"]["key1"]
    'val1'

    >>> parse_parameters(opts)["key4"]
    'key4param'

    >>> parse_parameters(opts)["key5"]
    'key5param'

    """
    names = ()
    return _parse_parameters(names, options)


def _parse_parameters(names, options):
    """Parse parameters, with a tuple of names for keyring context"""
    for key in options.keys():
        val = options[key]
        if isinstance(val, dict):
            val = _parse_parameters(names + (key,), val)
        if isinstance(val, string_types) and len(val) > 1 and val[0] == "^":
            # Decode a secret from the keystore
            val = val[1:]
            service = ".".join(names) or "_"
            if service == "resilient":
                # Special case, becuase of the way we parse commandlines, treat this as root
                service = "_"
            logger.debug("keyring get('%s', '%s')", service, val)
            val = keyring.get_password(service, val)

        if isinstance(val, string_types) and val.startswith(constants.PROTECTED_SECRET_PREFIX):
            config_name = val[1:]

            if helpers.protected_secret_exists(config_name, constants.PATH_SECRETS_DIR, constants.PATH_JWK_FILE):

                protected_secret = helpers.get_protected_secret(config_name, constants.PATH_SECRETS_DIR, constants.PATH_JWK_FILE)

                val = protected_secret if protected_secret else helpers.get_config_from_env(config_name)

            else:
                val = helpers.get_config_from_env(config_name)

        options[key] = val

    return options



# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.
import pytest
import unittest
import mock
from resilient.bin.res_keyring import KeyringUtils


class TestKeyringUtils(unittest.TestCase):
    mocked_resilient_class = mock.Mock()
    mocked_resilient_get_config = mock.Mock()

    @mock.patch("resilient.get_config_file", mocked_resilient_get_config)
    def test_run(self):
        mocked_get_config = mock.Mock()
        mocked_get_config = self.mocked_resilient_class.get_config_file.return_value

        self.mocked_resilient_get_config.return_value = "~/tmp"
        keyring_util = KeyringUtils()
        #
        # Not file there
        #
        assert keyring_util
        assert not keyring_util.config






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2021. All Rights Reserved.

"""Common paths used in tests"""

import os

SHARED_MOCK_DATA_DIR = os.path.dirname(os.path.realpath(__file__))
TESTS_DIR = os.path.dirname(SHARED_MOCK_DATA_DIR)

TEST_TEMP_DIR = os.path.join(TESTS_DIR, "test_temp")

MOCK_RESPONSES_DIR = os.path.join(SHARED_MOCK_DATA_DIR, "mock_responses")
MOCK_SECRETS_DIR = os.path.join(SHARED_MOCK_DATA_DIR, "mock_secrets")
